/*
 * rlog.c - Hold log data in output buffer for reading from /proc/rlogd
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/smp_lock.h>
#include <linux/console.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/interrupt.h>			/* For in_interrupt() */
#include <linux/slab.h>
#include <asm/uaccess.h>

#include <revirt/rlog.h>

// Might be nice to make this a tunable kernel parameter
//#define REPLAY_LOG_BUF_LEN	(1 << 22) /* This must be a power of two */
//#define REPLAY_LOG_BUF_MASK	(REPLAY_LOG_BUF_LEN-1)

/*
 * Write protection for these should be provided by revirt_logbuf_write_lock
 */
int revirt_logwait=0;
struct task_struct * revirt_waiting_tsk = NULL;


/*
 * revirt_logbuf_lock protects revirt_log_buf, revirt_log_start, 
 * revirt_log_end, revirt_read_start
 * 
 * revirt_logbuf_write_lock makes sure that writes are atomic; reads
 * may be interleaved with writes if necessary.
 */
DECLARE_MUTEX(revirt_logbuf_write_sem);
DECLARE_MUTEX(revirt_logbuf_sem);
//static spinlock_t revirt_logbuf_write_lock = SPIN_LOCK_UNLOCKED;
//static spinlock_t revirt_logbuf_lock = SPIN_LOCK_UNLOCKED;
DECLARE_WAIT_QUEUE_HEAD(revirt_log_read_wait);
DECLARE_WAIT_QUEUE_HEAD(revirt_log_write_wait);

//static char revirt_log_buf[REPLAY_LOG_BUF_LEN];
static char * revirt_log_buf=NULL;
static int revirt_log_buf_order=0;
#define REPLAY_LOG_BUF(idx) (revirt_log_buf[(idx) & revirt_options.logbuf.mask])

/*
 * The indices into revirt_log_buf are not constrained to REPLAY_LOG_BUF_LEN 
 * -- they must be masked before subscripting
 */
static unsigned long revirt_log_start;	/* Index into revirt_log_buf: next char to be read by revirtlog() */
static unsigned long revirt_lastread;	/* Index into revirt_log_buf: beginning of last data read.  (Not safe to throw away)*/
static unsigned long revirt_log_end;	/* Index into revirt_log_buf: most-recently-written-char + 1 */

#ifdef REPLAY_UNIQID
static unsigned int revirt_log_id=0;

int rlog_resetid(unsigned int i) {
	revirt_log_id=i;
	return revirt_log_id;
}
#endif

/*
 * Call to allocate or resize the log buf.  ztrace_setopt calls this with
 * realloc=1 -- i.e., only allocate if the buffer is already allocated.
 */
int revirt_alloc_log_buf(int realloc) {
	char * t;
	unsigned int l, order, pages, i;
	if(realloc && !revirt_log_buf)
		return 0;
	printk("revirt_alloc_log_buf: trying to allocate %d bytes\n", 
			revirt_options.logbuf.size);
	l=revirt_options.logbuf.size;
	pages = l / PAGE_SIZE;
	for(i=0; i<32; i++) {
		if((1UL << i) >= pages)
			break;
	}
	order = i;
	pages = 1UL << order;
	if(pages * PAGE_SIZE > l) {
		l = pages * PAGE_SIZE;
		revirt_options.logbuf.size = l;
		revirt_options.logbuf.mask = (l-1);
		printk("revirt_alloc_log_buf: changed to (%d %d %d %x)", 
				order,
				pages,
				revirt_options.logbuf.size,
				revirt_options.logbuf.mask);
	}
	/*
	 * With gcc 3.2, for some reason the allocator doesn't like 4
	 * meg allocations anymore... use __get_free_pages instead.
	 */
	printk("Trying to allocate pages order %d\n", order);
	t=(char *)__get_free_pages(GFP_KERNEL, order);
	if(!t)
		return -ENOMEM;
	/*
	 * FIXME - mild race condition
	 */
	if(revirt_log_buf) {
		printk("Warning: resizing logbuf is untested\n");
		/*
		 * This is poor naming; __get_free_pages returns a ptr to the
		 * actual page; __free_pages wants a struct page_struct *
		 * though; free_pages (no __) takes a ptr.
		 */
		free_pages((unsigned long)revirt_log_buf, 
				revirt_log_buf_order);
	}
	revirt_log_buf=t;
	revirt_log_buf_order = order;
	return 0;
}

/*
 * These are for the 'consumer' side of the equation -- i.e., /proc/rlog
 */
int revirt_rlog_open(void) {
	if(!revirt_log_buf) {
		return revirt_alloc_log_buf(0);
	}
	return 0;
}

int revirt_rlog_release(void) {
	return 0;
}

int revirt_rlog_read(char * buf, size_t len) {
	unsigned long i;
	//char c;
	int error;
	unsigned long read_end, real_start, real_end;

	error = -EINVAL;
	if (!buf || len < 0)
		goto out;
	error = 0;
	if (!len)
		goto out;

	if(!revirt_log_buf)  {
		error=-EIO;
		goto out;
	}
	/* 
	 * FIXME -- tocttou?
	 */
	error = verify_area(VERIFY_WRITE,buf,len);
	if (error)
		goto out;

	/*
	 * Note to self: waits until 2nd cond != 0 or signal pending
	 *
	 * Isn't this a race condition?
	 */
	wait_event_interruptible(revirt_log_read_wait, 
			    (revirt_log_end - revirt_log_start));
	if (error)
	      goto out;

	if(signal_pending(current)) {
		error=-ERESTARTSYS;
		goto out;
	}

	i = 0;

	if(down_interruptible(&revirt_logbuf_sem)<0) {
		error=-EINTR;
		goto out;
	}

	revirt_lastread = revirt_log_start;

	if(revirt_log_start > revirt_log_end) {
		printk("revirtlog: end before start: [%lx,%lx)\n",
			revirt_log_start, revirt_log_end);
		BUG();
	}
	/*
	 * Find the last "virtual index": either fill
	 * request, or whatever we have
	 */
	read_end = revirt_log_start + len;
	if(read_end > revirt_log_end)
		read_end = revirt_log_end;

	/*
	 * Now find the actual index from the virtual indices...
	 */

	real_start = revirt_log_start & revirt_options.logbuf.mask;
	real_end = read_end & revirt_options.logbuf.mask;

//	printk("revirtlog: start:0x%x end:0x%x readend:0x%x real_s:%x real_e:%x\n", 
	//		revirt_log_start, revirt_log_end, read_end,
//		real_start, real_end);
	/*
	 * If it's contiguious, do one copy; else, do
	 * two copies.
	 */
	if(real_end > real_start) {
		copy_to_user(buf, 
			 revirt_log_buf + real_start, 
			 real_end-real_start);
	} else {
		unsigned long d1, s1, l1, d2, s2, l2;
		s1 = (unsigned long)revirt_log_buf + real_start;
		l1 = revirt_options.logbuf.size - real_start;
		s2 = (unsigned long)revirt_log_buf;
		l2 = real_end;
		d1 = (unsigned long)buf;
		d2 = (unsigned long)buf + l1;
	//	printk("revirtlog: (%x,%x,%x) (%x,%x,%x)\n", 
	//		d1, s1, l1, d2, s2, l2);
		copy_to_user((void *)d1, (void *)s1, l1);
		copy_to_user((void *)d2, (void *)s2, l2);
	}
	i=read_end - revirt_log_start;
	revirt_log_start = read_end;
	up(&revirt_logbuf_sem);
	
	wake_up(&revirt_log_write_wait);
	error = i;
    out:
	return error;
}

unsigned int revirt_rlog_poll(void) {
	unsigned int error=0;
	if(down_interruptible(&revirt_logbuf_sem)<0)
		return -EINTR;
	error = revirt_log_end - revirt_log_start;
	up(&revirt_logbuf_sem);
	return error;
}

/*
 * Here's the producer side of the equation
 */
static void emit_rlog_char(char c)
{
	if(down_interruptible(&revirt_logbuf_sem)<0)
		return;
	REPLAY_LOG_BUF(revirt_log_end) = c;
	revirt_log_end++;
	/*
	 * Now that we're dealing with gigabytes, the log indices
	 * may overflow.  Rather than trying to deal with 64-bit indices,
	 * we detect overflow and compensate.  This means we need to be
	 * very careful about renewing our information before unlocking...
	 */
	if(revirt_log_end & 0x80000000) {
		revirt_log_end = (revirt_log_end & revirt_options.logbuf.mask) +
					revirt_options.logbuf.size;
		revirt_log_start = revirt_log_start & revirt_options.logbuf.mask;
		revirt_lastread = revirt_lastread & revirt_options.logbuf.mask;
		printk("rlog: index overflows, reset to [%lx,%lx)\n",
				revirt_log_start, revirt_log_end);
	}


	up(&revirt_logbuf_sem);
	/*
	 * This condition should never happen; the write lock combined with 
	 * the condition wait should guarantee that.
	 */
	if (revirt_log_end - revirt_log_start > revirt_options.logbuf.size) {
	   printk(KERN_EMERG "Replay Log overflow!!! Attempting to call schedule...\n");
	   schedule();
	}
	/*
	 * Just because we've sent the data to the rlogd process doesn't mean
	 * it's safely on disk.  We assume, however, that when rlogd calls
	 * read(/proc/rlog) again, all previously read data is on the disk.
	 * This warning detects the condition of overwriting data that has
	 * been sent to rlogd but we're not sure had made it to the disk yet.
	 */
	if (revirt_log_end - revirt_lastread > revirt_options.logbuf.size) {
	   printk("WARNING: Overwriting unsync'd log data\n");
	   revirt_lastread=revirt_log_start;
	}
}

/*
 * Must be called with revirt_logbuf_sem held!
 */
static int emit_log_chunk(char * b, int len) {
	int printed_len=0;
	unsigned long istart, iend, real_start, real_end, mid;
#if 0
	int i;
	for(i=0; i<len; i++, b++)  {
		emit_rlog_char(*b);
		printed_len++;
	}
#else
	if(len > revirt_options.logbuf.size)
		return 0;
	istart = revirt_log_end;
	iend = istart + len;  // This is the next free byte, not last written

	real_start = istart & revirt_options.logbuf.mask;
	real_end = iend & revirt_options.logbuf.mask;

	/*
	 * Normal case -- write contiguous chunk
	 */
	if(real_end > real_start) {
		// memcpy(dest, src, len)
		memcpy(revirt_log_buf+real_start, b, len);
	} else {
		/*
		 * Write straddles boundary: write two chunks
		 */
		mid = revirt_options.logbuf.size - real_start;
		memcpy(revirt_log_buf+real_start, b, mid);
		memcpy(revirt_log_buf, b+mid, real_end);
	}

	revirt_log_end = iend;
	printed_len = len;

	/*
	 * Reset log indices in case of a roll-over
	 */
	if(revirt_log_end & 0x80000000) {
		revirt_log_end = (revirt_log_end & revirt_options.logbuf.mask) +
					revirt_options.logbuf.size;
		revirt_log_start = revirt_log_start & revirt_options.logbuf.mask;
		revirt_lastread = revirt_lastread & revirt_options.logbuf.mask;
		printk("rlog: index overflows, reset to [%lx,%lx)\n",
				revirt_log_start, revirt_log_end);
	}

	/*
	 * This condition should never happen; the write lock combined with 
	 * the condition wait should guarantee that.
	 */
	if (revirt_log_end - revirt_log_start > revirt_options.logbuf.size) {
	   printk(KERN_EMERG "Replay Log overflow!!! Attempting to call schedule...\n");
	   schedule();
	}
	/*
	 * Just because we've sent the data to the rlogd process doesn't mean
	 * it's safely on disk.  We assume, however, that when rlogd calls
	 * read(/proc/rlog) again, all previously read data is on the disk.
	 * This warning detects the condition of overwriting data that has
	 * been sent to rlogd but we're not sure had made it to the disk yet.
	 */
	if (revirt_log_end - revirt_lastread > revirt_options.logbuf.size) {
	   printk("WARNING: Overwriting unsync'd log data\n");
	   revirt_lastread=revirt_log_start;
	}

#endif
	return printed_len;
}

/*
 * This is revirt_log_write. It can be called from any context.  
 * We want it to work.
 * We want to guarantee atomic writes of records, so this should be
 * with an entire record to write.
 */
asmlinkage int revirt_writev_log(int type, const struct iovec * v, int n);

asmlinkage int revirt_write_log(char * buf, int type, int len) {
	struct iovec v = { buf, len };
	return revirt_writev_log(type, &v, 1);
}

/*
 * If a logged process is blocked waiting for rlogd to empty the buffer,
 * we must not schedule another logged process.
 */
asmlinkage int revirt_log_can_schedule(void) {
	return !revirt_logwait;
}

static void revirt_flush_if_full(int type) { 
	if((revirt_log_end - revirt_log_start > revirt_options.logbuf.size / 4)
		 || (type == ENTRY_EXIT)) {
		//printk("revirtwrite: about to wake up revirt_log_read_wait\n");
		wake_up_interruptible(&revirt_log_read_wait);
	}
}

asmlinkage int revirt_writev_log(int type, const struct iovec * v, int n)
{
	int error;
	int printed_len=0, j,  len;
	struct rlog_header h;

	if(!revirt_log_buf) {
		error= revirt_alloc_log_buf(0);
		if(error)
			goto out;
	}

	for(len=j=0; j<n; j++)
		len+=v[j].iov_len;

	h.type=type;
	h.len=len;
#ifdef REPLAY_UNIQID
	h.id=revirt_log_id;
	revirt_log_id++;
	if(revirt_log_id==0) {
		printk("rlog: Warning: revirt_log_id rolled over\n");
	}
#endif

#if 0
	/*
	 * This is from the printk code. Is this necessary?  
	 */
	if (oops_in_progress) {
		/* If a crash is occurring, make sure we can't deadlock */
		spin_lock_init(&revirt_logbuf_write_lock);
	}
#endif
	
	len+=sizeof(h);

	/*
	 * In fact, if we did the space checking around emit_char() instead of
	 * here, we could handle lengths larger than the buffer size;
	 * per'aps we can add that later, but as we plan on the log being
	 * in the order of megabytes, it't not important right now.
	 */
	if(len > revirt_options.logbuf.size) {
		printk("revirtwrite: tried to write log of size %d, limit %d\n",
			len, revirt_options.logbuf.size);
		while(1);
	}

	/* FIXME -- hey, these aren't the right locks... we need semaphores */
	if(down_interruptible(&revirt_logbuf_write_sem)<0) {
		error=-EINTR;
		goto out;
	}

	while(revirt_log_end - revirt_log_start + len > revirt_options.logbuf.size) {
	    revirt_logwait=1;
	    revirt_waiting_tsk = current;
	    printk("revirtwrite: waiting for enough room to write data: s:%lx e:%lx l:%x LEN:%x\n", revirt_log_start, revirt_log_end, len, revirt_options.logbuf.size);
	    revirt_flush_if_full(type);
	    wait_event(revirt_log_write_wait, 
	        (!(revirt_log_end - revirt_log_start + len > revirt_options.logbuf.size)));
	    revirt_logwait=0;
	    revirt_waiting_tsk = NULL;
	}

	if(down_interruptible(&revirt_logbuf_sem)<0) {
		error=-EINTR;
		goto out_up;
	}
	printed_len += emit_log_chunk((char *)&h, sizeof(h));
	for(j=0; j<n; j++)
	   printed_len+=error=emit_log_chunk((char *)v[j].iov_base, 
			   			v[j].iov_len);
	up(&revirt_logbuf_sem);

	/*
	 * For optimization, only wake up rlogd when log has reached some
	 * watermark, and when a logged process exits.  We don't want to
	 * schedule after a signal, however, so don't wake up then.
	 *
	 * Note, this means that REPLAY_SCHED should never be set when
	 * writing type ENTRY_EXIT.
	 */
	if((printed_len && !(current->revirt_flags & REPLAY_SCHED))) 
		revirt_flush_if_full(type);
	error=len;
       out_up:
	up(&revirt_logbuf_write_sem);
       out:
	return error;
}
EXPORT_SYMBOL(revirt_write_log);
