/*
 * $Id: umkbd.c,v 1.1.1.1 2003/05/27 21:45:37 root Exp $
 */

/*
 * This is not a driver but a simulator reacting on in/out requests to
 * I/O addresses in range 0x60 - 0x6f.
 * see resource allocation in include/asm-i386/keyboard.h
 */
/* FIXME: for the moment there is only a keyboard, no mouse. */

/* In include/asm-i386/keyboard.h four functions are defined which access the
 * memory region 0x60-0x6f using inb and outb
 * Keyboard controller registers are defined in drivers/sbus/char/pcikbd.h
 * relative to pcikbd_iobase (which is probably 0x60, where is this set?).
 * According to documentation from IBM:
 * KBD_STATUS_REG =        0x64 (when reading from this address)
 * KBD_CNTL_REG   =        0x64 (when writing to this address)
 * KBD_DATA_REG   =        0x60 (both reading/writing to/from this address)
 *
 * Looks to me as though keyboard/mouse do everything bytewise (no inw/outw etc.)
 */
#include <linux/assert.h>
#include <asm/lotol.h>
#include <asm/lotol_fdesc.h>
#include <asm-i386/fcntl.h>
#include <asm-i386/signal.h>
/*#include <asm-i386/keyboard.h>*/
#include <asm/ptrace.h>
#include <linux/sched.h>

#ifndef wait_queue_head_t /* ##tg## -> needed for Debian woody linux/pc_keyb.h */
#define wait_queue_head_t     struct wait_queue *
#endif

#include <linux/pc_keyb.h>

/* see arch/i386-linux/kernel/irq.h */
/* and drivers/char/pc_keyb.c */
#define KBD_IRQ	1
#define AUX_IRQ	12

/* these addresses are used hardcoded in arch/i386-linux/kernel/entry.c */
#define UMKBD_STATUS_REG 0x0064
#define UMKBD_CNTL_REG 0x0064
#define UMKBD_DATA_REG 0x0060

/*
 * Status Register Bits
 */
#define KBD_STAT_PERR 		0x80 /* Parity error */
#define KBD_STAT_GTO 		0x40 /* General receive/xmit timeout */
#define KBD_STAT_MOUSE_OBF	0x20 /* Mouse output buffer full */
#define KBD_STAT_UNLOCKED	0x10 /* Zero if keyboard locked */
#define KBD_STAT_CMD		0x08 /* Last write was a command write (0=data) */
#define KBD_STAT_SELFTEST	0x04 /* Self test successful */
#define KBD_STAT_IBF 		0x02 /* Keyboard input buffer full */
#define KBD_STAT_OBF 		0x01 /* Keyboard output buffer full */

/*
 * Keyboard Controller Commands
 */
#define KBD_CCMD_READ_MODE	0x20 /* Read mode bits */
#define KBD_CCMD_WRITE_MODE	0x60 /* Write mode bits */
#define KBD_CCMD_GET_VERSION	0xA1 /* Get controller version */
#define KBD_CCMD_RSV1		0xA4 /* reserved */
#define KBD_CCMD_RSV2		0xA5 /* reserved */
#define KBD_CCMD_RSV3		0xA6 /* reserved */
#define KBD_CCMD_MOUSE_DISABLE	0xA7 /* Disable mouse interface */
#define KBD_CCMD_MOUSE_ENABLE	0xA8 /* Enable mouse interface */
#define KBD_CCMD_TEST_MOUSE	0xA9 /* Mouse interface test */
#define KBD_CCMD_SELF_TEST	0xAA /* Controller self test */
#define KBD_CCMD_KBD_TEST	0xAB /* Keyboard interface test */
#define KBD_CCMD_KBD_DISABLE	0xAD /* Keyboard interface disable */
#define KBD_CCMD_KBD_ENABLE	0xAE /* Keyboard interface enable */
#define KBD_CCMD_RDIN		0xC0 /* read input port */
#define KBD_CCMD_PLINH		0xC2 /* poll input port high */
#define KBD_CCMD_PLINL		0xC3 /* poll input port low */
#define KBD_CCMD_RDOUT		0xD0 /* read output port */
#define KBD_CCMD_WROUT		0xD1 /* read output port */
#define KBD_CCMD_WRKOR		0xD2 /* write keyboard output register */
#define KBD_CCMD_WRITE_AUX_OBUF	0xD3 /* write auxiliary device output register*/
#define KBD_CCMD_WRITE_MOUSE	0xD4 /* Write the following byte to the mouse */
#define KBD_CCMD_RDTST		0xE0 /* read test input port */
#define KBD_CCMD_PLSE		0xF0 /* pulse: bits 3-0 select which bit */

/* Keyboard replies */
#define UMKBD_REPLY_POR		0xAA /* Power on reset */
#define UMKBD_REPLY_ACK		0xFA /* Command ACK */
#define UMKBD_REPLY_RESEND	0xFE /* Command NACK, send the cmd again */

/* Keyboard commands */
#define UMKBD_CMD_SET_LEDS	0xED /* Set keyboard leds */
#define UMKBD_CMD_SET_RATE	0xF3 /* Set typematic rate */
#define UMKBD_CMD_ENABLE	0xF4 /* Enable scanning */
#define UMKBD_CMD_DISABLE	0xF5 /* Disable scanning */
#define UMKBD_CMD_RESET		0xFF /* Reset */

static unsigned char modereg = 0; /* controller mode register */
static unsigned char umkbd_led_status = 0;
static unsigned char umkbd_mouse_enabled = 0;

/* some keyboard commands are followed by data, so remember the command */
static unsigned char umkbd_keyboard_command = 0;
/* some controller commands are followed by data, so remember the command */
static unsigned char umkbd_controller_command = 0;

#define DEBUG_INOUT_KBD 0

#if DEBUG_INOUT_KBD
#define dprintk(fmt, args...) printk(fmt, ##args)
#else
#define dprintk(fmt, args...)
#endif


static unsigned short umkbd_buf[256];
static int umkbd_count = 0;
static int umkbd_head = 0;
static int umkbd_tail = 0;


static void
umkbd_push(unsigned short cmd)
{
	extern void call_isa_irq(int irq);

	if (umkbd_count < sizeof(umkbd_buf)) {
		umkbd_buf[umkbd_head] = cmd;
		umkbd_count++;
		umkbd_head = (umkbd_head + 1) % sizeof(umkbd_buf);

		if (umkbd_count == 1) {
			switch (cmd >> 8) {
			case 1: /* keyboard */
				call_isa_irq(KBD_IRQ);
				break;
			case 2: /* keyboard */
				call_isa_irq(AUX_IRQ);
				break;
			default:
				assert(0);
			}
		}
	}
}

static unsigned char
umkbd_pop(void)
{
	extern void call_isa_irq(int irq);
	unsigned char value;

	assert(0 < umkbd_count);

	/* disable keyboard interrupt while updating umkbd_count! FIXME */
	value = (unsigned char) umkbd_buf[umkbd_tail];
	umkbd_count--;
	umkbd_tail = (umkbd_tail + 1) % sizeof(umkbd_buf);

	if (0 < umkbd_count) {
		switch (umkbd_buf[umkbd_tail] >> 8) {
		case 1:	/* keyboard */
			call_isa_irq(KBD_IRQ);
			break;
		case 2: /* keyboard */
			call_isa_irq(AUX_IRQ);
			break;
		default:
			assert(0);
		}
	}

	return value;
}

static unsigned char
incode_to_trcode(unsigned char incode)
{
	static unsigned char trcode[128] = {
	     /*    0     1     2     3     4     5     6     7 */
	        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 00 */
		0x0e, 0x0f, 0x1c, 0x00, 0x00, 0x1c, 0x00, 0x00, /* 01 */
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* 02 */
		0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, /* 03 */
		0x39, 0x02, 0x28, 0x04, 0x05, 0x06, 0x08, 0x28, /* 04 */
		0x0a, 0x0b, 0x09, 0x0d, 0x33, 0x0c, 0x34, 0x35, /* 05 */
		0x0b, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, /* 06 */
		0x09, 0x0a, 0x27, 0x27, 0x33, 0x0d, 0x34, 0x35, /* 07 */
		0x03, 0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22, /* 10 */
		0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18, /* 11 */
		0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11, /* 12 */
		0x2d, 0x15, 0x2c, 0x1a, 0x2b, 0x1b, 0x07, 0x0c, /* 13 */
		0x29, 0x1e, 0x30, 0x2e, 0x20, 0x12, 0x21, 0x22, /* 14 */
		0x23, 0x17, 0x24, 0x25, 0x26, 0x32, 0x31, 0x18, /* 15 */
		0x19, 0x10, 0x13, 0x1f, 0x14, 0x16, 0x2f, 0x11, /* 16 */
		0x2d, 0x15, 0x2c, 0x1a, 0x2b, 0x1b, 0x29, 0x00, /* 17 */
	};
	
	return trcode[incode];
}

void
umkbd_interrupt(void)
{
  	unsigned short cmd;
	/* Changed to read ascii from stdin and convert it to scancode -MAB */
	static unsigned short lshiftdn = 0x12a, lshiftup = 0x1aa;
	static unsigned short lctrldn = 0x11d, lctrlup = 0x19d;
	unsigned char incode, trcode;
	int ret, shift = 0, ctrl = 0;
	//char msg[8];

	dprintk("---umkbd--- %s\n", __FUNCTION__);


	for (;;) {
		ret = lotol_read(0, &incode, sizeof(incode));
		if (ret <= 0) {
			break;
		}
		assert(ret == sizeof(incode));

		if (incode < 27) {
		        ctrl = 1;
			incode = incode + 'A' - 1;
		}
		if ( (incode >= '!' && incode <= '&') ||
		     (incode >= '(' && incode <= '+') || (incode == ':') ||
		     (incode == '<') || (incode >= '>' && incode <= 'Z') ||
		     (incode == '^') || (incode == '_') ||
		     (incode >= '{' && incode <= '~') ) {
		        shift = 1;
		}

		trcode = incode_to_trcode(incode);
		/*
		sprintf(msg, " %c %03u ", incode, incode);
		do {
		        ret = lotol_write(1, msg, strlen(msg));
		} while (ret < 0 && (errno == LOTOL_EINTR ||
				     errno == LOTOL_EAGAIN));
		*/
		if (shift) {
		        umkbd_push(lshiftdn);
		}
		if (ctrl) {
		        umkbd_push(lctrldn);
		}

		cmd = trcode | (1 << 8); /* Key down */
		if (umkbd_mouse_enabled || (cmd >> 8) != 2) {
			umkbd_push(cmd);
		}

		cmd = trcode | (1 << 8) | 0x80; /* Key up */
		if (umkbd_mouse_enabled || (cmd >> 8) != 2) {
			umkbd_push(cmd);
		}

		if (shift) {
		        umkbd_push(lshiftup);
		}
		if (ctrl) {
		        umkbd_push(lctrlup);
		}
	}
}


/*
 * keyboard received a command in previous write, now receives data for this
 * command (e.g. for setting leds).
 * FIXME: This might need to be fixed to really do something, to make the driver
 * happy.
 */
void
do_keyboard_command_data(unsigned char value)
{
	dprintk("%s: call()\n", __FUNCTION__);

	switch (umkbd_keyboard_command) {
	case UMKBD_CMD_SET_LEDS:
		umkbd_led_status = value;
		umkbd_push((1 << 8) | UMKBD_REPLY_ACK);
		break;

	default:
		assert(0);
	}

	umkbd_keyboard_command = 0;

	dprintk("%s: return\n", __FUNCTION__);
}

void do_keyboard_command(unsigned char value)
{
	dprintk("%s: call(value=0x%02x)\n", __FUNCTION__, value);

	switch (value) {
	case UMKBD_CMD_DISABLE:
		umkbd_push((1 << 8) | UMKBD_REPLY_ACK);
		break;

	case UMKBD_CMD_ENABLE:
		umkbd_push((1 << 8) | UMKBD_REPLY_ACK);
		break;

	case UMKBD_CMD_SET_RATE:
		umkbd_push((1 << 8) | UMKBD_REPLY_ACK);
		break;

	case UMKBD_CMD_SET_LEDS:
		umkbd_keyboard_command = UMKBD_CMD_SET_LEDS;
		umkbd_push((1 << 8) | UMKBD_REPLY_ACK);
		break;

	case UMKBD_CMD_RESET:
		/*
		 * Now reset the keyboard, which will take a while.
		 * When we are done, put UMKBD_REPLY_POR into buffer,
		 * since driver is waiting for this.
		 */
		umkbd_push((1 << 8) | UMKBD_REPLY_ACK);
		umkbd_push((1 << 8) | UMKBD_REPLY_POR);
		break;

	default:
		printk("WARNING: %s: value=0x%02x\n", __FUNCTION__, value);
		break;
	}

	dprintk("%s: return\n", __FUNCTION__);
}


/*
 *
 */
static void
do_controller_command_data(unsigned char value)
{
	dprintk("%s call(value=0x%02x\n", __FUNCTION__, value);

	switch (umkbd_controller_command) {
	case KBD_CCMD_WRITE_MODE:
		/* enable/disable interrupts FIXME */
		modereg = value;
		break;

	case KBD_CCMD_WRITE_AUX_OBUF:
		umkbd_push((2 << 8) | value);
		break;

	case KBD_CCMD_WRITE_MOUSE:	/* write to aux device FIXME */
		break;
	}

	umkbd_controller_command = 0;

	dprintk("%s: return\n", __FUNCTION__);
}

static void
do_controller_command(unsigned char value)
{
	dprintk("%s call(value=0x%02x\n", __FUNCTION__, value);

	/*
	 * Commands used for keyboard initialization in initialize_keyboard()
	 * in drivers/char/pc_keyb.c
	 */
	switch (value) {
	case KBD_CCMD_SELF_TEST:
		/*
		 * Bit 0 of the controller status register is set to 1 upon
		 * completion of self test. Linux driver doesn't check this.
		 */
		umkbd_push((1 << 8) | 0x55);
		break;

	case KBD_CCMD_KBD_TEST:
		umkbd_push((1 << 8) | 0x00);
		break;

	case KBD_CCMD_KBD_ENABLE:
		break;

	case KBD_CCMD_WRITE_MODE:	/* write mode register */
		umkbd_controller_command = value;
		break;

	case KBD_CCMD_READ_MODE:	/* read mode register */
		umkbd_push((1 << 8) | modereg);
		break;

	case KBD_CCMD_WRITE_AUX_OBUF:	/* write to output buffer as if */
					/* initiated by the auxiliary device */
		umkbd_controller_command = value;
		break;

	case KBD_CCMD_MOUSE_DISABLE:	/* disable mouse */
		umkbd_mouse_enabled = 0;
		break;

	case KBD_CCMD_MOUSE_ENABLE:	/* enable mouse */
		umkbd_mouse_enabled = 1;
		break;

	case KBD_CCMD_WRITE_MOUSE:	/* write to aux device FIXME */
		umkbd_controller_command = value;
		break;

	case KBD_CCMD_TEST_MOUSE: /* Other commands: */
	case KBD_CCMD_KBD_DISABLE:
	case KBD_CCMD_RDIN :
	case KBD_CCMD_PLINH:
	case KBD_CCMD_PLINL:
	case KBD_CCMD_RDOUT:
	case KBD_CCMD_WROUT:
	case KBD_CCMD_WRKOR:
	case KBD_CCMD_RDTST:
	case KBD_CCMD_PLSE :
		/* a7 */
		printk("WARNING: %s: value=0x%02x\n", __FUNCTION__, value);
		break;
	default:
		assert(0);
		break;
	}

	dprintk("%s: return\n", __FUNCTION__);
}

/*
 * device to controller to OS (OS read)
 */
unsigned char
umkbd_inb(unsigned short addr)
{
	unsigned char value;

	dprintk("%s: call(addr=0x%04x)\n", __FUNCTION__, addr);

	switch (addr) {
	case UMKBD_DATA_REG:
		if (0 < umkbd_count) {
			value = umkbd_pop();
		} else {
			value = 0x00;
		}
		break;

	case UMKBD_STATUS_REG:
		value = 0;

		/* parity error? */
		/* value |= KBD_STAT_PERR; */

		/* timing error? */
		/* value |= KBD_STAT_GTO; */

		/* mouse output? */
		if (1 <= umkbd_count && (umkbd_buf[umkbd_tail] >> 8) == 2) {
			value |= KBD_STAT_MOUSE_OBF;
		}

		/* keyboard unlocked? */
		value |= KBD_STAT_UNLOCKED;

		/* command or data FIXME VOSSI */
		/* value |= KBD_STAT_CMD; */

		/* system flag FIXME VOSSI */
		/* value |= KBD_STAT_SELFTEST; */

		/* data in input buffer? */
		/* value |= KBD_STAT_IBF; */

		/* output buffer full? */
		if (1 <= umkbd_count) {
			value |= KBD_STAT_OBF;
		}
		break;

	default:
		panic("%s: addr=0x%04x\n", __FUNCTION__, addr);
		value = 0x00;
		break;
	}

	dprintk("%s: return(value=0x%02x)\n", __FUNCTION__, value);
	return value;
}

/*
 * os to controller to device
 */
void
umkbd_outb(unsigned char value, unsigned short addr)
{
	dprintk("%s: call(value=0x%02x, addr=0x04x)\n",
			__FUNCTION__, value, addr);

	switch (addr) {
	case UMKBD_DATA_REG:
		if (umkbd_controller_command) {
			do_controller_command_data(value);

		} else if (umkbd_keyboard_command) {
			do_keyboard_command_data(value);

		} else {
			do_keyboard_command(value);
		}
		break;

	case UMKBD_CNTL_REG:
		do_controller_command(value);
		break;

	default:
		/* FIXME VOSSI */
		/* panic("%s: addr=0x%04x\n", __FUNCTION__, addr); */
		break;
	}

	dprintk("%s: return\n", __FUNCTION__);
}

/*
 * initialize keyboard hardware
 */
void
umkbd_init(void)
{
	int flags;
	int ret;

	dprintk("%s: call()\n", __FUNCTION__);

	/* Changed to operate on stdin instead of FD_KEYBOARD -MAB */

	/*
	 * send signal if data available
	 * don't block if no data is available
	 */
	flags = lotol_fcntl(0, LOTOL_F_GETFL, 0);
	flags |= LOTOL_O_ASYNC;
	flags |= LOTOL_O_NONBLOCK;
	ret = lotol_fcntl(0, LOTOL_F_SETFL, flags);
	if (ret < 0) {
		panic("lotol_fcntl: F_SETFL\n");
	}

	/*
	 * set receiver of signal
	 */
	ret = lotol_fcntl(0, LOTOL_F_SETOWN, lotol_getpid());
	if (ret < 0) {
		panic("lotol_fcntl: F_SETOWN\n");
	}
}
