/*
 * Network driver for linux system.
 * Copyright Volkmar Sieh / University Erlangen/Nueremberg
 */

static const char *version =
	"UMN Version 1.0 by University Erlangen-Nuremberg\n";

#include <linux/assert.h>

#include <linux/init.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>

#include <linux/if_arp.h>	/* For ARPHRD_ETHER */

#include <asm/lotol.h>
#include <asm/lotol_fdesc.h>

#define ETHER_ADDR_LEN	6


struct umn_private {
	int fd;
	int myself;
	struct {
		unsigned char mac[ETHER_ADDR_LEN];;
		long addr;
		int port;
	} host[16];
	struct net_device_stats stats;
};

static void umn_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	struct net_device *dev = (struct net_device *) dev_id;
	struct umn_private *umn = (struct umn_private *) dev->priv;

	static char buf[65536];
	int len;
	struct sk_buff *skb;
	lotol_sigset_t nmask;
	lotol_sigset_t omask;
	int ret;

	/* disable (real) interrupts */
	lotol_sigemptyset(&nmask);
	lotol_sigaddset(&nmask, LOTOL_SIGALRM);
	lotol_sigaddset(&nmask, LOTOL_SIGHUP);
	lotol_sigaddset(&nmask, LOTOL_SIGIO);
	lotol_sigaddset(&nmask, LOTOL_SIGURG);
	lotol_sigaddset(&nmask, LOTOL_SIGVTALRM);
	ret = lotol_sigprocmask(LOTOL_SIG_SETMASK, &nmask, &omask);
	assert(0 <= ret);

	for (;;) {
		do {
		        len = lotol_read(umn->fd, buf, sizeof(buf));
			assert(0 < len
			    || errno == LOTOL_EAGAIN
			    || errno == LOTOL_ECONNREFUSED
			    || errno == LOTOL_EIO);
		} while (len < 0
		      && (errno == LOTOL_ECONNREFUSED
		       || errno == LOTOL_EIO));
		assert(0 < len || errno == LOTOL_EAGAIN);
		if (len < 0)
			/* no more packets available */
			break;

		skb = dev_alloc_skb(len + 2);
		if (! skb) {
			printk(KERN_INFO "%s: Memory squeeze, deferring packet.\n",
					dev->name);
			goto err;
		}

		skb->dev = dev;
		skb_reserve(skb, 2);			/* 16 byte align */
		skb_put(skb, len);			/* make room */
		eth_copy_and_sum(skb, buf, len, 0);
		skb->protocol = eth_type_trans(skb, dev);
		netif_rx(skb);

		umn->stats.rx_bytes += len;
		umn->stats.rx_packets++;
	}

err:
	/* restore (real) interrupts */
	ret = lotol_sigprocmask(LOTOL_SIG_SETMASK, &omask, (lotol_sigset_t *) 0);
	assert(0 <= ret);
}

static int umn_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct umn_private *umn = (struct umn_private *) dev->priv;
	struct ethhdr *hdr;
	int ret;


	assert(skb != NULL);
	assert(sizeof(*hdr) <= skb->len);

#if 0	/* FIXME HSHOEXER */
	dev->tbusy = 1;
#endif
	
	hdr = (struct ethhdr *) skb->data;
	ret = lotol_write(umn->fd, skb->data, skb->len);

	if (ret < 0)
	  printk("TUN_TAP: umn_xmit: errno: %d\n", errno);
	
	dev_kfree_skb(skb);

#if 0	/* FIXME HSHOEXER */
	dev->tbusy = 0;
#endif

	umn->stats.tx_bytes += skb->len;
	umn->stats.tx_packets++;

	return 0;
}

static struct net_device_stats *umn_get_stats(struct net_device *dev)
{
	struct umn_private *up = (struct umn_private *) dev->priv;

	return &up->stats;
}

static void umn_set_multicast(struct net_device *dev)
{
	/* FIXME */
}

static int umn_open(struct net_device *dev)
{
	if (request_irq(dev->irq, umn_interrupt, SA_SHIRQ,
			dev->name, dev)) {
		printk(KERN_ERR "%s: Can't get irq %d\n",
				dev->name, dev->irq);
		return -EAGAIN;
	}
#if 0	/* FIXME HSHOEXER */
	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;
#endif
	return 0;
}

static int umn_close(struct net_device *dev)
{
#if 0	/* FIXME HSHOEXER */
	dev->start = 0;
	dev->tbusy = 1;
#endif
	free_irq(dev->irq, dev);
	return 0;
}

int __init tuntap_probe(struct net_device *dev)
{
	static unsigned version_printed = 0;

	struct umn_private *umn;
	int i;
	int fd;
	char c;
	char offset;
	int ret;

	if (version_printed++ == 0)
		printk(version);

	/*
	 * allocate private memory
	 */
	umn = kmalloc(sizeof(struct umn_private), GFP_KERNEL);
	if (umn == NULL) {
		printk("%s: unable to get memory for dev->priv.\n", dev->name);
		return ENOMEM;
	}

	assert(strlen(dev->name) == 4);
	offset = dev->name[3] - '0';
	assert(offset <= 7);

	/*
	 * read network description from file
	 */
	fd = FD_TUNTAP_CONF;
	ret = lotol_lseek(fd, 0, 0);
	if (ret < 0) {
		kfree(umn);
		return ENXIO;
	}
	ret = lotol_read(fd, &c, 1);

	if (ret < 0) {
		kfree(umn);
		return ENXIO;
	}
	/* read `myself' */
	umn->myself = 0;
	if (ret != 1) c = '\0';
	while (('0' <= c && c <= '9')
	     || ('A' <= c && c <= 'F')
	     || ('a' <= c && c <= 'f')) {
		umn->myself *= 16;
		if ('0' <= c && c <= '9')
			umn->myself += c - '0';
		else if ('A' <= c && c <= 'F')
			umn->myself += c - 'A' + 10;
		else if ('a' <= c && c <= 'f')
			umn->myself += c - 'a' + 10;
		ret = lotol_read(fd, &c, 1);
		if (ret != 1) c = '\0';
	}
	/* read mac address / ip address/port */
	for (i = 0; i < sizeof(umn->host) / sizeof(umn->host[0]); i++) {
		int j;

		/* read mac address */
		for (j = 0; j < ETHER_ADDR_LEN; j++) {
			/* read digits */
			umn->host[i].mac[j] = 0;
			ret = lotol_read(fd, &c, 1);
			if (ret != 1) c = '\0';
			while (('0' <= c && c <= '9')
			    || ('A' <= c && c <= 'F')
			    || ('a' <= c && c <= 'f')) {
				umn->host[i].mac[j] *= 16;
				if ('0' <= c && c <= '9')
					umn->host[i].mac[j] += c - '0';
				else if ('A' <= c && c <= 'F')
					umn->host[i].mac[j] += c - 'A' + 10;
				else if ('a' <= c && c <= 'f')
					umn->host[i].mac[j] += c - 'a' + 10;
				ret = lotol_read(fd, &c, 1);
				if (ret != 1) c = '\0';
			}
		}

		/* read ip address */
		umn->host[i].addr = 0;
		ret = lotol_read(fd, &c, 1);
		if (ret != 1) c = '\0';
		while (('0' <= c && c <= '9')
		    || ('A' <= c && c <= 'F')
		    || ('a' <= c && c <= 'f')) {
			umn->host[i].addr *= 16;
			if ('0' <= c && c <= '9')
				umn->host[i].addr += c - '0';
			else if ('A' <= c && c <= 'F')
				umn->host[i].addr += c - 'A' + 10;
			else if ('a' <= c && c <= 'f')
				umn->host[i].addr += c - 'a' + 10;
			ret = lotol_read(fd, &c, 1);
			if (ret != 1) c = '\0';
		}

		/* read ip port */
		umn->host[i].port = 0;
		ret = lotol_read(fd, &c, 1);
		if (ret != 1) c = '\0';
		while (('0' <= c && c <= '9')
		    || ('A' <= c && c <= 'F')
		    || ('a' <= c && c <= 'f')) {
			umn->host[i].port *= 16;
			if ('0' <= c && c <= '9')
				umn->host[i].port += c - '0';
			else if ('A' <= c && c <= 'F')
				umn->host[i].port += c - 'A' + 10;
			else if ('a' <= c && c <= 'f')
				umn->host[i].port += c - 'a' + 10;
			ret = lotol_read(fd, &c, 1);
			if (ret != 1) c = '\0';
		}
	}

	/*
	 * allocate resources
	 */
	dev->irq = 11;	/* FIXME VOSSI */
	dev->base_addr = (int) dev;	/* use some address */
	dev->dma = 0;

	/*
	 * init device
	 */

	/* TUN_TAP */
	
	umn->fd = FD_TUNTAP;

	ret = lotol_fcntl(umn->fd, LOTOL_F_SETOWN, lotol_getpid());
	if (ret < 0) {
		printk("%s: can't set owner\n", dev->name);
		/* (void) lotol_close(umn->fd); */
		return ENXIO;
	}


	memset(&umn->stats, 0, sizeof(umn->stats));

	for (i = 0; i < ETHER_ADDR_LEN; i++) {
		dev->dev_addr[i] = umn->host[umn->myself].mac[i];
	}
	dev->priv = umn;
	dev->open = umn_open;
	dev->stop = umn_close;
	dev->hard_start_xmit = umn_xmit;
	dev->get_stats = umn_get_stats;
	dev->set_multicast_list = umn_set_multicast;

	ether_setup(dev);

	printk("%s: found (using IRQ %d, MAC", dev->name, dev->irq);
	for (i = 0; i < ETHER_ADDR_LEN; i++) {
		printk("%c%2.2x", i == 0 ? ' ' : ':',
				umn->host[umn->myself].mac[i]);
	}
	printk(", port 0x%x)\n", umn->host[umn->myself].port);

	return 0;
}

void umeth_interrupt(void)
{
	extern void call_isa_irq(unsigned int);

	call_isa_irq(11);	/* FIXME VOSSI */
}

void umeth_init(void)
{
}
