/*
 * $Id: umsmp.c,v 1.1.1.1 2003/05/27 21:45:53 root Exp $
 * vim:tabstop=8:shiftwidth=8:textwidth=72
 */

#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "external_constants.h"
#include "lotol_fdesc.h"

#include "lilo.h"

/* XXX too many magic numbers, more define should be used... */

#define PA_MPF	0xF5000
#define PA_MPC	0xF1000
#define MPF (MEM_BASE + PA_MPF)
#define MPC (MEM_BASE + PA_MPC)

/* Some usefull definitions from include/i386/mpspec.h */
struct intel_mp_floating
{
	char mpf_signature[4];		/* "_MP_" 			*/
	unsigned long mpf_physptr;	/* Configuration table address	*/
	unsigned char mpf_length;	/* Our length (paragraphs)	*/
	unsigned char mpf_specification;/* Specification version	*/
	unsigned char mpf_checksum;	/* Checksum (makes sum 0)	*/
	unsigned char mpf_feature1;	/* Standard or configuration ? 	*/
	unsigned char mpf_feature2;	/* Bit7 set for IMCR|PIC	*/
	unsigned char mpf_feature3;	/* Unused (0)			*/
	unsigned char mpf_feature4;	/* Unused (0)			*/
	unsigned char mpf_feature5;	/* Unused (0)			*/
};

struct mp_config_table
{
	char mpc_signature[4];
#define MPC_SIGNATURE "PCMP"
	unsigned short mpc_length;	/* Size of table */
	char  mpc_spec;			/* 0x01 */
	char  mpc_checksum;
	char  mpc_oem[8];
	char  mpc_productid[12];
	unsigned long mpc_oemptr;	/* 0 if not present */
	unsigned short mpc_oemsize;	/* 0 if not present */
	unsigned short mpc_oemcount;
	unsigned long mpc_lapic;	/* APIC address */
	unsigned long reserved;
};

/* Followed by entries */

#define	MP_PROCESSOR	0
#define	MP_BUS		1
#define	MP_IOAPIC	2
#define	MP_INTSRC	3
#define	MP_LINTSRC	4
#define	MP_TRANSLATION  192  /* Used by IBM NUMA-Q to describe node locality */

struct mpc_config_processor
{
	unsigned char mpc_type;
	unsigned char mpc_apicid;	/* Local APIC number */
	unsigned char mpc_apicver;	/* Its versions */
	unsigned char mpc_cpuflag;
#define CPU_ENABLED		1	/* Processor is available */
#define CPU_BOOTPROCESSOR	2	/* Processor is the BP */
	unsigned long mpc_cpufeature;		
#define CPU_STEPPING_MASK 0x0F
#define CPU_MODEL_MASK	0xF0
#define CPU_FAMILY_MASK	0xF00
	unsigned long mpc_featureflag;	/* CPUID feature value */
	unsigned long mpc_reserved[2];
};

struct mpc_config_ioapic
{
	unsigned char mpc_type;
	unsigned char mpc_apicid;
	unsigned char mpc_apicver;
	unsigned char mpc_flags;
#define MPC_APIC_USABLE         0x01
	unsigned long mpc_apicaddr;
};

struct mpc_config_intsrc
{
	unsigned char mpc_type;
	unsigned char mpc_irqtype;
	unsigned short mpc_irqflag;
	unsigned char mpc_srcbus;
	unsigned char mpc_srcbusirq;
	unsigned char mpc_dstapic;
	unsigned char mpc_dstirq;
};

enum mp_irq_source_types {
	mp_INT = 0,
	mp_NMI = 1,
	mp_SMI = 2,
	mp_ExtINT = 3
};

#define MP_IRQDIR_DEFAULT	0
#define MP_IRQDIR_HIGH		1
#define MP_IRQDIR_LOW		3

/* Calculate checksum */
int checksum(unsigned char *p, int len)
{
	int sum = 0;
	while (len--)
		sum += *p++;

	sum &= 0xFF;

	return (0x100 - sum);
}

/* Builds a MP configuration table (normally done by the BIOS) */
void
umsmp_init(int ncpu)
{
#ifdef SMP
	int i;
	struct intel_mp_floating *mpf;
	struct mp_config_table *mpc;
	struct mpc_config_processor *mcpu;
	struct mpc_config_intsrc *ints;
	struct mpc_config_ioapic *ioapic;

	mpf = (struct intel_mp_floating*)MPF;

	sprintf(mpf->mpf_signature, "_MP_");
	mpf->mpf_physptr = PA_MPC;
	mpf->mpf_length = 1;
	mpf->mpf_specification = 4;
	mpf->mpf_feature1 = 0x0;
	mpf->mpf_feature2 = 0x80;
	mpf->mpf_feature3 = 0x0;
	mpf->mpf_feature4 = 0x0;
	mpf->mpf_feature5 = 0x0;

	mpf->mpf_checksum = checksum((unsigned char*)mpf, 
			sizeof(struct intel_mp_floating));

	mpc = (struct mp_config_table*)MPC;

	sprintf(mpc->mpc_signature, MPC_SIGNATURE);
	mpc->mpc_length = sizeof(struct mp_config_table) 
		+ ncpu * sizeof(struct mpc_config_processor) 
		+ 16 * sizeof(struct mpc_config_intsrc)
		+ sizeof(struct mpc_config_ioapic);

	mpc->mpc_spec = 4;
	sprintf(mpc->mpc_oem, "IMMD3");
	sprintf(mpc->mpc_productid, "LOTOL");
	mpc->mpc_oemptr = 0;
	mpc->mpc_oemsize = 0;
	mpc->mpc_oemcount = 0;
	mpc->mpc_lapic = PA_LOCAL_APIC;
	mpc->reserved = 0;

	mcpu = (struct mpc_config_processor*)(MPC 
			+ sizeof(struct mp_config_table));

	for (i = 0; i < ncpu; i++, mcpu++) {
		mcpu->mpc_type = MP_PROCESSOR;
		mcpu->mpc_apicid = i;
		mcpu->mpc_apicver = 0x1;
		mcpu->mpc_cpuflag = (CPU_ENABLED | (!i ? CPU_BOOTPROCESSOR : 0));
		assert(( (i == 0) && mcpu->mpc_cpuflag == 3) ||
			 ((0 < i) && mcpu->mpc_cpuflag == 1));

		mcpu->mpc_cpufeature = 0x0510;
		mcpu->mpc_featureflag = 0x20;
		mcpu->mpc_reserved[0] = 0;
		mcpu->mpc_reserved[1] = 0;
	}

	ints = (struct mpc_config_intsrc *)(MPC 
			+ sizeof(struct mp_config_table)
			+ ncpu * sizeof(struct mpc_config_processor));

	for (i = 0; i < 16; i++) {
		if (i == 2) continue;	/* IRQ2 is never connected */
		ints->mpc_type = MP_INTSRC;
		ints->mpc_irqtype = mp_INT;
		ints->mpc_irqflag = 0;
		ints->mpc_srcbus = 0;
		ints->mpc_dstapic = 0x01;
		ints->mpc_srcbusirq = i;
		ints->mpc_dstirq = i ? i : 2;	/* IRQ0 to INTIN 2*/
		ints++;
	}

	ints->mpc_type = MP_INTSRC;	/* 8259A to INTIN0 */
	ints->mpc_irqtype = mp_ExtINT;
	ints->mpc_irqflag = 0;
	ints->mpc_srcbus = 0;
	ints->mpc_srcbusirq = 0;
	ints->mpc_dstirq = 0;

	ioapic = (struct mpc_config_ioapic *)(MPC 
			+ sizeof(struct mp_config_table) 
			+ ncpu * sizeof(struct mpc_config_processor) 
			+ sizeof(struct mpc_config_intsrc) * 16);
	
	ioapic->mpc_type = MP_IOAPIC;
	ioapic->mpc_apicid = ncpu;
	ioapic->mpc_apicver = 0x01;
	ioapic->mpc_flags = MPC_APIC_USABLE;
	ioapic->mpc_apicaddr = PA_IO_APIC;

	mpc->mpc_checksum = checksum((unsigned char*)mpc, 
			sizeof(struct mp_config_table) 
			+ ncpu * sizeof(struct mpc_config_processor) 
			+ sizeof(struct mpc_config_intsrc) * 16
			+ sizeof(struct mpc_config_ioapic));
#endif	/* SMP */
}
