/*
 * ipt_user_replay.c
 *
 * Will read data from iprecorddata and iprecordsize in current directory and
 * try to construct a TCP stream from it.
 * Will give error/warning messages to stderr if it finds something wrong with
 * thd data. Will also send a lot of debugging output to stdout if DEBUG is
 * defined.
 * 
 */

/* Hack to get ntohl etc working */
#define _LINUX_BYTEORDER_GENERIC_H	1

#define _GNU_SOURCE

#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/netfilter.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netpacket/packet.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>
#include "libipq.h"

#define DEBUG


volatile int term = 0;

FILE *fp1, *fp2;

char *sizes;

unsigned lost_bytes, repeated_bytes;

struct {
  int sock;
  unsigned char port[2];
  unsigned seq;
} connection[64];


void
int_handler(int sig)
{
  switch(sig) {
  case SIGPIPE:
	fprintf(stderr, "SIGNAL: Caught SIGPIPE, ignoring.\n");
	break;
  default:
    fprintf(stderr, "SIGNAL: going down on signal %d : %s\n", sig,
	    strsignal(sig));
    term = 1;
  }
}


void
dump(char *string, unsigned char *buf, int bufsize)
{
  int i;

  fprintf(stdout, "%s: \n", string);
  for (i = 0; i < bufsize && i < 100; i++) {
    fprintf(stdout, "%02x ", buf[i]);
  }
  fprintf(stdout, "\n");
}


void
openfile(void)
{
  unsigned size;
  if( (fp1 = fopen("./iprecordsize", "r")) == NULL ) {
    fprintf(stderr, "FILEIO: Unable to open file ./iprecordsize\n");
    term = 1;
  }
  
  fseek(fp1, 0, SEEK_END);
  size = ftell(fp1);
  fseek(fp1, 0, SEEK_SET); 
  sizes = (char *)malloc(size);
  if( (fread((void *)sizes, 1, size, fp1)) != size ){
    fprintf(stderr, "FILEIO: Error in reading from file ./iprecordsize\n");
    term = 1;
  }
  fclose(fp1);
  sizes[size-1] = '\0';

  if( (fp2 = fopen("./iprecorddata", "r")) == NULL ) {
    fprintf(stderr, "FILEIO: Unable to open file ./iprecorddata\n");
    term = 1;
  }
}


void
closefile(void)
{
  fclose(fp2);
}


unsigned char *
readrecord(int recordsize)
{
  unsigned char *buf = (unsigned char *)malloc(recordsize);
  if( (fread((void *)buf, 1, recordsize, fp2)) 
      != recordsize ){
    fprintf(stderr, "FILEIO: Error in reding from file ./iprecorddata\n");
    term = 1;
  }
  return buf;
}


/*
 * port2sock manages the ports in the connection array.
 * action = 0 : check whether the port exists.
 *          1 : add the port to the connection array.
 *          2 : delete the port from the connection array.
 */
int
port2sock(unsigned char *src_port, int action)
{
  int i;
  static unsigned char zero[2] = { 0, 0 };


  switch (action) {

  case 0:
    for( i = 0; i < 64; i++) {
      if( memcmp(src_port, connection[i].port , 2) == 0 ) {
#ifdef DEBUG
	fprintf(stdout, "SOCK: port2sock - check\n");
	fprintf(stdout, "\t src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
#endif
	return i;
      }
    }
#ifdef DEBUG
    fprintf(stdout, "SOCK: port2sock - check\n");
    fprintf(stdout, "\t src_port: %02x %02x Check Failed.\n", src_port[0],
	    src_port[1]);
#endif
    return -1;

  case 1:
    for( i = 0; i < 64; i++) {
      if( memcmp(connection[i].port, zero, 2) == 0 ) {
	memcpy(connection[i].port, src_port, 2);
#ifdef DEBUG
	fprintf(stdout, "SOCK: port2sock - add\n");
	fprintf(stdout, "\t src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
#endif
	return i;
      }
    }
#ifdef DEBUG
    fprintf(stdout, "SOCK: port2sock - add\n");
    fprintf(stdout, "\t src_port: %02x %02x Add Failed.\n", src_port[0],
	    src_port[1]);
#endif
    return -1;    

  case 2:
    for( i = 0; i < 64; i++) {
      if( memcmp(src_port, connection[i].port, 2) == 0 ) {
	memcpy(connection[i].port, zero, 2);
#ifdef DEBUG
	fprintf(stdout, "SOCK: port2sock - del\n");
	fprintf(stdout, "\t src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
#endif
	return i;
      }
    }
#ifdef DEBUG
    fprintf(stdout, "SOCK: port2sock - del\n");
    fprintf(stdout, "\t src_port: %02x %02x Del Failed.\n", src_port[0],
	    src_port[1]);
#endif
    return -1;
  }
  
  return -1;
}


/*
 * seqmismatch prints warning messages and updates counters when sequence
 * numbers in packets are not as expected
 */
void
seqmismatch(unsigned expected, unsigned obtained)
{
  fprintf(stderr, "PARSE: WARNING: Sequence number mismatch. Expected"
	  " %u got %u.\n", expected, obtained);
#ifdef DEBUG
  fprintf(stdout, "PARSE: WARNING: Sequence number mismatch. Expected"
	  " %u got %u.\n", expected, obtained);
#endif
  if( expected < obtained ) {
    fprintf(stderr, "\tMissed %u bytes.\n", obtained - expected);
#ifdef DEBUG
    fprintf(stdout, "\tMissed %u bytes.\n", obtained - expected);
#endif
    lost_bytes += obtained - expected;
  }
  else {
    fprintf(stderr, "\tRepeated %u bytes.\n", expected - obtained);
#ifdef DEBUG
    fprintf(stdout, "\tRepeated %u bytes.\n", expected - obtained);
#endif
    repeated_bytes += expected - obtained;
  }
}


/*
 * parserecord understand the contents of the packet and does the needfull.
 * 
 * The destination_port should always be as expected as the recording function
 * takes care of this.
 * Assumptions are made about ihl, version, fragmentation and protocol fields
 * and a debugging message will be sent to stderr if they are broken.
 *
 * Optional TCP headers are not explicitely taken care of, but the correct TCP
 * data is found and sent using write.
 * Socket management is done when a syn, fin or res packet is encountered.
 * Otherwise if the packet contains TCP data, it is written to the appropriate
 * connection (if it exists). The pipe is always emptied by reading data from 
 * it before writing anything to it.
 */
void
parserecord(unsigned char *buf, int recordsize, unsigned char *xserver_ip)
{
  unsigned char control;
  int urg, ack, pus, res, syn, fin;
  unsigned char tcp_header_len;
  int tcp_header_length;
  int sockno;
  const int opt = 1;
  struct sockaddr_in server;
#ifdef DEBUG
  static int i = 0;
#endif
  const int junksize = 65536;
  unsigned char junk[junksize];
  int readret;
  int close_connection = 0;
  unsigned char zero_ip[4] = { 0, 0, 0, 0 };
  int data_size;
  unsigned seq_no;

  // Version + Header Length[1], Type of Service[1].
  static unsigned char misc1[2] = {
    0x45, 0
  };
  // Skip Length[2], Id[2].
  // Fragmentation offset[2] for no fragmentation.
  static unsigned char frag[2] = {
    0x40, 0
  };
  // Skip Time to live[1].
  // Protocol[1] for TCP.
  static unsigned char proto[1] = {
    0x06
  };
  // Skip Checksum[2].
  // Skip IP address of Source[4].
  // Skip IP address of destination[4].
  // Skip Source port[2].
  // Destination port[2]. (6000)
  static unsigned char dst_port[2] = {
    0x17, 0x70
  };
  // Skip Sequence no[4], Acknowledgement no[4].

  if ( memcmp(buf + 22, dst_port, 2) != 0 ) {
    // Panic.
    fprintf(stderr, "PARSE: Error in parsing headers\n");
    dump("error", buf, recordsize);
    term = 1;
    return;
  }

  if ( memcmp(buf, misc1, 2) != 0
       || memcmp(buf + 6, frag, 2) != 0
       || memcmp(buf + 9, proto, 1) != 0
       ) {
    // Concern.
    fprintf(stderr, "PARSE: WARNING: Trouble in parsing headers\n");
    dump("warning", buf, recordsize);
    return;
  }

  // Extract Offset[4/8]
  tcp_header_len = (buf[32] & 0xf0) >> 4;
  tcp_header_length = tcp_header_len*4;
  data_size = recordsize-20-tcp_header_length;

  // Skip Reserved[6/8]
  // Extract Control Flags[6/8]
  control = buf[33];
  urg = (control & 0x20) >> 5;
  ack = (control & 0x10) >> 4;
  pus = (control & 0x08) >> 3;
  res = (control & 0x04) >> 2;
  syn = (control & 0x02) >> 1;
  fin = control & 0x01;

  seq_no = buf[24+0]*256u*256u*256u + buf[24+1]*256u*256u + buf[24+2]*256u 
           + buf[24+3];


  if(syn) {

#ifdef DEBUG
    fprintf(stdout, "PARSE: Starting syn\n");
    dump("syn", buf, recordsize);
#endif

    // Close socket if connection exists.
    sockno = port2sock(buf+20, 0);
    if ( sockno != -1 ) {
      close(connection[sockno].sock);
      port2sock(buf+20, 2);
    }

    // Prepare for new connection.
    sockno = port2sock(buf+20, 1);
    if ( sockno == -1 ) {
      fprintf(stderr, "PARSE: Exceeded max sockets\n");
      term = 1;
      return;
    }

    // Set sequence number to that of next packet to be sent on this
    // connection.
    connection[sockno].seq = seq_no + 1;

    // Create socket.
    connection[sockno].sock = socket( AF_INET, SOCK_STREAM, 0 );
    if (connection[sockno].sock == -1) {
      fprintf(stderr, "PARSE: Error in opening socket\n");
      term = 1;
    }
    
    // Configure the socket to send data immediately.
    if( setsockopt(connection[sockno].sock, IPPROTO_TCP, TCP_NODELAY,
		   (void *)&opt, sizeof(int)) == -1) {
      fprintf(stderr, "PARSE: Error in setting socket options\n");
      term = 1;
    }

    // Connect socket.
    server.sin_family = AF_INET;

    if( memcmp(xserver_ip, zero_ip, 4) == 0 )
      memcpy(xserver_ip, buf+16, 4);

    memcpy(&server.sin_addr.s_addr, xserver_ip, 4);
    memcpy(&server.sin_port, buf+22, 2);

#ifdef DEBUG
    //fprintf(stdout, "PARSE: Before connect\n");
#endif
    if (connect(connection[sockno].sock, (struct sockaddr *) &server,
		sizeof server) == -1) {
      fprintf(stderr, "PARSE: Error in connecting socket\n");
      fprintf(stderr, "\t Errno: %d\n", errno);
      term = 1;
    }

    //fflush(NULL);

    // Make the socket non-blocking for recv calls.
    if( fcntl(connection[sockno].sock, F_SETFL, O_NONBLOCK) == -1 ) {
      fprintf(stderr, "PARSE: Error in making socket non-blocking\n");
      term = 1;
    }

#ifdef DEBUG
    //fprintf(stdout, "PARSE: Ending syn\n");
#endif
  }
  else if(fin) {

#ifdef DEBUG
    fprintf(stdout, "PARSE: Starting fin\n");
    dump("fin", buf, recordsize);
#endif

    // Close socket if connection exists.
    sockno = port2sock(buf+20, 0);
    if ( sockno != -1 ) {
      // Check if sequence number is as expected.
      if( connection[sockno].seq != seq_no ) {
	seqmismatch(connection[sockno].seq, seq_no);
      }
      close(connection[sockno].sock);
      port2sock(buf+20, 2);
    }
    else {
#ifdef DEBUG
      fprintf(stdout, "PARSE: No connection to close.\n");
#endif
    }

  }
  else if(res) {

#ifdef DEBUG
    fprintf(stdout, "PARSE: Connection reset by Peer\n");
    dump("reset", buf, recordsize);
#endif

    // Close socket if connection exists.
    sockno = port2sock(buf+20, 0);
    if ( sockno != -1 ) {
      // Check if sequence number is as expected.
      if( connection[sockno].seq != seq_no ) {
	seqmismatch(connection[sockno].seq, seq_no);
      }
      close(connection[sockno].sock);
      port2sock(buf+20, 2);
    }
    else {
#ifdef DEBUG
      fprintf(stdout, "PARSE: No connection to reset.\n");
#endif
    }

  }
  else {
	  
    if( data_size > 0 ) {

      sockno = port2sock(buf+20, 0);
      if ( sockno != -1 ) {
	do {
	  readret = read(connection[sockno].sock, junk, junksize);
#ifdef DEBUG
	  fprintf(stdout, "PARSE: Read returned  %d.\n", readret);
#endif
	  if( readret == 0 ) {
	    fprintf(stderr,"PARSE: Sender closed connection!\n");
#ifdef DEBUG
	    fprintf(stdout,"PARSE: Sender closed connection!\n");
#endif
	    close(connection[sockno].sock);
	    port2sock(buf+20, 2);
	    close_connection = 1;
	  }
	} while( readret == junksize );
#ifdef DEBUG
	fprintf(stdout, "PARSE: Before writing packet %d.\n", i);
	dump("write", buf, recordsize);
#endif
	if( close_connection == 0 ) {

	  // Check if sequence number is as expected.
	  if( connection[sockno].seq != seq_no ) {
	    seqmismatch(connection[sockno].seq, seq_no);
	  }

	  if(write(connection[sockno].sock, buf+20+tcp_header_length,
		   data_size) == -1) {
	    if( errno == EPIPE ) {
	      fprintf(stderr,
		      "PARSE: Got EPIPE in write, closing connection\n");
#ifdef DEBUG
	      fprintf(stdout,
		      "PARSE: Got EPIPE in write, closing connection\n");
#endif
	      close(connection[sockno].sock);
	      port2sock(buf+20, 2);
	    }
	    else {
	      fprintf(stderr, "PARSE: Error in write\n");
	      term = 1;
	    }
	  }
	  //fflush(NULL);

	  // Set sequence number to that of next packet to be sent on this
	  // connection.
	  connection[sockno].seq = seq_no + data_size; 

	}
	else {
	  fprintf(stdout, "PARSE: Connection closed. Not writing packet.\n");
	  dump("write", buf, recordsize);
	}
      }
      else {
	fprintf(stdout, "PARSE: No connection. Not writing packet.\n");
	dump("write", buf, recordsize);
      }
    }
    else {
#ifdef DEBUG
      fprintf(stdout, "PARSE: Packet without data portion.\n");
      dump("ignore", buf, recordsize);
#endif
    }
    
  }
#ifdef DEBUG
  fprintf(stdout, "PARSE: Record size: %4d   Data size: %4d\n", recordsize,
	  data_size);
  fprintf(stdout, "\n");
  i++;
#endif
  //fflush(NULL);
}



int
main(int argc, char *argv[] )
{
  struct sigaction sa;
  int ret;
  char *token;
  int recordsize;
  unsigned char *buffer;
  unsigned long int sleep_time;
  unsigned char ip[4] = { 0, 0, 0, 0 };


  if (argc < 2) {
    fprintf(stderr, "Usage: %s <Sleep Time in microseconds> [<XServerIP>]\n", 
	    argv[0]);
    exit(1);
  }
  
  sleep_time = strtoul(argv[1], NULL, 10);

#ifdef DEBUG
  fprintf(stdout, "%s: Sleep Time in microsec %lu\n", argv[0], sleep_time);
#endif

  if (argc == 3) {
    ret = inet_aton(argv[2], (struct in_addr *)ip );
    assert(0 <= ret);
#ifdef DEBUG
    fprintf(stdout, "Sending packets to XServer on IP: ");
    fprintf(stdout, "%02x %02x %02x %02x\n", ip[0], ip[1], ip[2], ip[3]);
#endif
  }


  sa.sa_handler = int_handler;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  ret = sigaction(SIGINT, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGQUIT, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGHUP, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGPIPE, &sa, (struct sigaction *) 0);
  assert(0 <= ret);


  openfile();

  if( ( token = strtok(sizes, " ") ) != NULL ) {
    recordsize = atoi(token);
    buffer = readrecord(recordsize);
    parserecord(buffer, recordsize, ip);
    free(buffer);
  }

  while(!term) {
    if( ( token = strtok(NULL, " ") ) != NULL ) {
      recordsize = atoi(token);
      buffer = readrecord(recordsize);
      usleep(sleep_time);
      parserecord(buffer, recordsize, ip);
      free(buffer);
    }
    else
      term = 1;
  }

  closefile();

  if( lost_bytes > 0 ) {
    fprintf(stderr, "Lost %u bytes totally.\n", lost_bytes);
#ifdef DEBUG
    fprintf(stdout, "Lost %u bytes totally.\n", lost_bytes);
#endif
  }
  if( repeated_bytes > 0 ) {
    fprintf(stderr, "Repeated %u bytes totally.\n", repeated_bytes);
#ifdef DEBUG
    fprintf(stdout, "Repeated %u bytes totally.\n", repeated_bytes);
#endif
  }
  
  return 0;
}
