/*
 * net.c
 *
 * Uses SOCK_RAW to pass packets to and from UMLinux Guest.
 *
 * During logging (-l), it passes all packets going to the guest, and accepts
 * any packets coming from it.
 *
 * During replay (-r) it sends no packets to the guest. When the guest sends
 * packets destined for the XServer, it interprets them and opens new TCP
 * connections to the XServer and sends the same data on those connections.
 * If the packets are coming too fast, it will temporarily stop the zr process
 * which sends it the packets.
 * PROBLEM: Some packets do get dropped, if they come too fast. This will be
 * detected due to sequence number mismatch, and a message will be sent to
 * stderr.
 *
 */

/* Hack to get ntohl etc working */
#define _LINUX_BYTEORDER_GENERIC_H	1

#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netpacket/packet.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <netdb.h>
#include <fcntl.h>

//#define DEBUG


char *progname;

volatile int term = 0;
struct ifreq saved_ifr;

int eth;
int myself;

unsigned lost_bytes, repeated_bytes;

struct {
  int sock;
  unsigned char port[2];
  unsigned seq;
} connection[64];

struct {
  unsigned char mac[6];
  unsigned long addr;
  unsigned short port;
} host[16];


void
int_handler(int sig)
{
  switch(sig) {
    case SIGPIPE:
	fprintf(stderr, "Caught sigpipe. ignoring\n");
	break;
    default:
	fprintf(stderr, "%s: going down on signal %d\n", progname, sig);
	term = 1;
  }
}


void
dump(char *string, unsigned char *buf, int bufsize)
{
  int i;

  fprintf(stdout, "%s: \n", string);
  for (i = 0; i < bufsize && i < 100; i++) {
    fprintf(stdout, "%02x ", buf[i]);
  }
  fprintf(stdout, "\n");
}


int
phys_open(void)
{
  int fd;
  struct ifreq ifr;
  struct sockaddr_ll addr;
  int ifindex;
  int ret;
  
  
  /* ================= */
  /* create raw socket */
  /* ================= */
  
  fd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
  if (fd < 0) {
    fprintf(stderr, "%s: socket(PF_PACKET, SOCK_RAW, ETH_P_ALL): %s.\n",
	    progname, strerror(errno));
    return -1;
  }
  
  /* =================== */
  /* get interface index */
  /* =================== */
  
  memset(&ifr, 0, sizeof(ifr));
  sprintf(ifr.ifr_name, "eth%d", eth);
  ret = ioctl(fd, SIOCGIFINDEX, &ifr);
  if (ret < 0) {
    fprintf(stderr, "%s: ioctl(..., SIOCGIFINDEX, ...): %s.\n",
	    progname, strerror(errno));
    (void) close(fd);
    return -1;
  }
  ifindex = ifr.ifr_ifindex;

  /* ================== */
  /* set promiscue mode */
  /* ================== */
  
  memset(&ifr, 0, sizeof(ifr));
  sprintf(ifr.ifr_name, "eth%d", eth);
  ret = ioctl(fd, SIOCGIFFLAGS, &ifr);
  if (ret < 0) {
    fprintf(stderr, "%s: ioctl(..., SIOCGIFFLAGS, ...): %s.\n",
	    progname, strerror(errno));
    (void) close(fd);
    return -1;
  }
  saved_ifr = ifr;
  ifr.ifr_flags |= IFF_PROMISC;
  ret = ioctl(fd, SIOCSIFFLAGS, &ifr);
  if (ret < 0) {
    fprintf(stderr, "%s: ioctl(..., SIOCSIFFLAGS, ...): %s.\n",
	    progname, strerror(errno));
    (void) close(fd);
    return -1;
  }
  
  /* =========== */
  /* bind socket */
  /* =========== */
  
  memset(&addr, 0, sizeof(addr));
  addr.sll_family = PF_PACKET;
  addr.sll_protocol = htons(ETH_P_ALL);
  addr.sll_ifindex = ifindex;
  ret = bind(fd, (struct sockaddr *) &addr, sizeof(addr));
  if (ret < 0) {
    fprintf(stderr, "%s: bind: %s.\n",
	    progname, strerror(errno));
    (void) close(fd);
    return -1;
  }

#ifdef DEBUG  
  fprintf(stdout, "ifindex: %d\n", ifindex);
  fprintf(stdout, "phys_open: fd = %d\n", fd);
#endif

  return fd;
}


int
phys_recv(int fd, unsigned char *buf, int bufsize)
{
  struct sockaddr_ll addr;
  int addrlen = sizeof(addr);
  int len;

  addrlen = sizeof(addr);  
  len = recvfrom(fd, buf, bufsize, 0, (struct sockaddr *) &addr, &addrlen);

  return len;
}


int
phys_send(int fd, char *buf, int bufsize)
{
  int ret;
  
  ret = sendto(fd, buf, bufsize, 0, (struct sockaddr *) 0, 0);

#ifdef DEBUG  
  fprintf(stdout, "phys_send: ret = %d\n", ret);
#endif

  return ret;
}


int
phys_close(int fd)
{
  int ret;
  
  if (fd < 0) {
    return 0;
  }

  ret = ioctl(fd, SIOCSIFFLAGS, &saved_ifr);
  if (ret < 0) {
    fprintf(stderr, "%s: ioctl(..., SIOCSIFFLAGS, ...): %s.\n",
	    progname, strerror(errno));
  }
  
  ret = close(fd);
  
#ifdef DEBUG
  fprintf(stdout, "phys_close: ret = %d\n", ret);
#endif

  return ret;
}


int
virt_open(void)
{
  int fd;
  struct sockaddr_in addr;
  int ret;
  
  fd = socket(PF_INET, SOCK_DGRAM, 0);
  if (fd < 0) {
    fprintf(stderr, "%s: socket(PF_INET, SOCK_DGRAM, 0): %s.\n",
	    progname, strerror(errno));
    return -1;
  }

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = INADDR_ANY;
  addr.sin_port = htons(host[myself].port);
  ret = bind(fd, (struct sockaddr *) &addr, sizeof(addr));
  if (ret < 0) {
    fprintf(stderr, "%s: bind: %s.\n",
	    progname, strerror(errno));
    (void) close(fd);
    return -1;
  }
  
#ifdef DEBUG
  fprintf(stdout, "virt_open: fd = %d\n", fd);
#endif

  return fd;
}


int
virt_recv(int fd, char *buf, int bufsize)
{
  struct sockaddr_in addr;
  int addrlen;
  int len;

  addrlen = sizeof(addr);
  len = recvfrom(fd, buf, bufsize, 0, (struct sockaddr *) &addr, &addrlen);
  
  return len;
}


int
virt_send(int fd, char *buf, int bufsize)
{
  struct sockaddr_in addr;
  int i;
  int ret;

  ret = bufsize;
  
  for (i = 0; i < 16; i++) {
    if (i == myself) continue;
    if (host[i].addr == 0 || host[i].port == 0) continue;
    
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(host[i].addr);
    addr.sin_port = htons(host[i].port);
    
    if (sendto(fd, buf, bufsize, 0, (struct sockaddr *)&addr, 
	       sizeof(addr)) != bufsize) {
      ret = -1;
    }
  }
  
  return ret;
}


int
virt_close(int fd)
{
  int ret;
  
  if (fd < 0) {
    return 0;
  }
  
  ret = close(fd);

#ifdef DEBUG
  fprintf(stdout, "virt_close: ret = %d\n", ret);
#endif

  return ret;
}


/* 
 * Return 0 for packets to guest IP/ARP or broadcast IP/MAC or unknown.
 * Return 1 otherwise (Packet to be dropped).
 */

int
to_guest_filter(unsigned char *buf, int bufsize, unsigned char *guest_ip,
		int filter)
{

  // Broadcast MAC address[6].
  static unsigned char broadcast[6] = {
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff
  };

  // MAC address for guest[6]. (HACK)
  static unsigned char guest_mac[6] = {
    0, 0xff, 0, 0, 0, 0
  };

  // Frame type for IP[2].
  static unsigned char frame_type_ip[2] = {
    0x08, 0
  };

  // Frame type for ARP[2].
  static unsigned char frame_type_arp[2] = {
    0x08, 0x06
  };

  switch(filter) {

  case 0: // Broadcast + All IP to guest + All ARP to guest + All non-IP/ARP.
    if( (memcmp(buf, broadcast, 6) == 0) ||
	( (memcmp(buf + 12, frame_type_ip, 2) == 0) && 
	  (memcmp(buf + 30, guest_ip, 4) == 0) ) ||
	( (memcmp(buf + 12, frame_type_arp, 2) == 0) && 
	  (memcmp(buf + 38, guest_ip, 4) == 0) ) ||
	( (memcmp(buf + 12, frame_type_ip, 2) != 0) && 
	  (memcmp(buf + 12, frame_type_arp, 2) != 0) ) ) {
      return 0;
    }
    break;

  case 1: // Broadcast + All non-IP + IP to guest
    if( (memcmp(buf, broadcast, 6) == 0) ||
	(memcmp(buf + 12, frame_type_ip, 2) != 0) ||
	(memcmp(buf + 30, guest_ip, 4) == 0) ) {
      return 0;
    }
    break;

  case 2: // HACK: Broadcast + All ETH to guest + All IP to guest
    if( (memcmp(buf, broadcast, 6) == 0) ||
	(memcmp(buf, guest_mac, 6) == 0) ||
	(memcmp(buf + 30, guest_ip, 4) == 0) ) {
      return 0;
    }
    break;

  case 3: // EVERYTHING
    return 0;
    break;

  default:
    fprintf(stderr, "%s: invalid filter option: %d.\n",
	    progname, filter);
    term = 1;
    break;
  };

  return 1;
}


/* 
 * Return 0 for packets to XServer IP on port 6000.
 * Return 1 otherwise (Packet to be dropped).
 */

int
to_xserver_filter(unsigned char *buf, int bufsize, unsigned char *xserver_ip)
{

  // Frame type for IP[2].
  static unsigned char frame_type_ip[2] = {
    0x08, 0
  };

  // XServer Destination port[2]. (6000)
  static unsigned char dst_port[2] = {
    0x17, 0x70
  };

  if(  (memcmp(buf + 12, frame_type_ip, 2) == 0) && 
       (memcmp(buf + 30, xserver_ip, 4) == 0) &&
      (memcmp(buf + 36, dst_port, 2) == 0) ) {
    return 0;
  }

  return 1;
}


static unsigned long
read_hex(int fd, int digits, char delim, char *file)
{
  unsigned long n;
  int d;
  char c;
  int ret;

  n = 0;
  for (d = 0; d < digits; d++) {
    ret = read(fd, &c, 1);
    if (ret < 0) {
      fprintf(stderr, "%s: read error in %s: %s.\n",
	      progname, file, strerror(errno));
      exit(1);
    } else if (ret == 0) {
      fprintf(stderr, "%s: read error in %s: %s.\n",
	      progname, file, "End-of-file");
    }
    n *= 16;
    if ('0' <= c && c <= '9')
      n += c - '0';
    else if ('A' <= c && c <= 'F')
      n += c - 'A' + 10;
    else if ('a' <= c && c <= 'f')
      n += c - 'a' + 10;
    else {
      fprintf(stderr, "%s: read error in %s: %s.\n",
	      progname, file, "No hex digit");
      exit(1);
    }
  }
  ret = read(fd, &c, 1); if (ret != 1) c = '\0';
  if (c != delim) {
    fprintf(stderr, "%s: read error in %s: %s.\n",
	    progname, file, "Bad delimiter");
    exit(1);
  }
  
  return n;
}


static void
config_read(void)
{
  char *filename;
  int fd;
  int i;
  
  eth = 0;	/* FIXME */
  filename = "eth0";
  fd = open(filename, O_RDONLY);
  if (fd < 0) {
    fprintf(stderr, "%s: can't open %s: %s.\n",
	    progname, filename, strerror(errno));
    exit(1);
  }
  
  /* read myself */
  myself = read_hex(fd, 2, '\n', filename);
  
  /* read mac / address:port table */
  for (i = 0; i < sizeof(host) / sizeof(host[0]); i++) {
    int j;
    
    for (j = 0; j < 6; j++) {
      host[i].mac[j] =
	read_hex(fd, 2, j == 5 ? ' ' : ':', filename);
    }
    host[i].addr = read_hex(fd, 8, ':', filename);
    host[i].port = read_hex(fd, 4, '\n', filename);
  }
  
  (void) close(fd);
}


/*
 * port2sock manages the ports in the connection array.
 * action = 0 : check whether the port exists.
 *          1 : add the port to the connection array.
 *          2 : delete the port from the connection array.
 */
int
port2sock(unsigned char *src_port, int action)
{
  int i;
  static unsigned char zero[2] = { 0, 0 };


  switch (action) {

  case 0:
    for( i = 0; i < 64; i++) {
      if( memcmp(src_port, connection[i].port , 2) == 0 ) {
#ifdef DEBUG
	fprintf(stdout, "SOCK: port2sock - check\n");
	fprintf(stdout, "\t src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
#endif
	return i;
      }
    }
#ifdef DEBUG
    fprintf(stdout, "SOCK: port2sock - check\n");
    fprintf(stdout, "\t src_port: %02x %02x Check Failed.\n", src_port[0],
	    src_port[1]);
#endif
    return -1;

  case 1:
    for( i = 0; i < 64; i++) {
      if( memcmp(connection[i].port, zero, 2) == 0 ) {
	memcpy(connection[i].port, src_port, 2);
#ifdef DEBUG
	fprintf(stdout, "SOCK: port2sock - add\n");
	fprintf(stdout, "\t src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
#endif
	return i;
      }
    }
#ifdef DEBUG
    fprintf(stdout, "SOCK: port2sock - add\n");
    fprintf(stdout, "\t src_port: %02x %02x Add Failed.\n", src_port[0],
	    src_port[1]);
#endif
    return -1;    

  case 2:
    for( i = 0; i < 64; i++) {
      if( memcmp(src_port, connection[i].port, 2) == 0 ) {
	memcpy(connection[i].port, zero, 2);
#ifdef DEBUG
	fprintf(stdout, "SOCK: port2sock - del\n");
	fprintf(stdout, "\t src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
#endif
	return i;
      }
    }
#ifdef DEBUG
    fprintf(stdout, "SOCK: port2sock - del\n");
    fprintf(stdout, "\t src_port: %02x %02x Del Failed.\n", src_port[0],
	    src_port[1]);
#endif
    return -1;
  }
  
  return -1;
}

/*
int
port2sock(unsigned char *src_port, int caller, int *close_socket)
{
  int i;
  static unsigned char map[64][2];
  static unsigned char zero[2] = { 0, 0 };

  switch(caller) {

  case 0: // Syn
    
    for( i = 0; i < 64; i++ )
      if( memcmp(src_port, map[i], 2) == 0 ) {
	fprintf(stderr, "%s: Trying to send syn from port with established"
		" connection\n", progname);
	fprintf(stderr, "\t: src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
	*close_socket = 1;
	break;
      }
    // FALLTHRU


    if( *close_socket != 1 )
      for( i = 0; i < 64; i++ )
	if( memcmp(map[i], zero, 2) == 0 ) {
	  memcpy(map[i], src_port, 2);
#ifdef DEBUG
	  fprintf(stderr, "%s: src_port: %02x %02x i: %d\n", progname,
		  src_port[0], src_port[1], i);
#endif
	  break;
	}
    break;

  case 1: // Fin
  case 2: // Res

    for( i = 0; i < 64; i++ )
      if( memcmp(src_port, map[i], 2) == 0 ) {
#ifdef DEBUG
	fprintf(stderr, "\t: src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
#endif
	memcpy(map[i], zero, 2);
	break;
      }
    break;
    
  case 3: // Data
       
    for( i = 0; i < 64; i++ )
      if( memcmp(src_port, map[i], 2) == 0 ) {
#ifdef DEBUG
	fprintf(stderr, "\t: src_port: %02x %02x i: %d\n", src_port[0],
		src_port[1], i);
#endif
	break;
      }
    break;

   case 4: // EPIPE
	for(i=0; i< 64; i++) {
		if(! memcmp(src_port, map[i], 2))
			break;
	}
	memcpy(map[i], zero, 2);
	break;
  default:
	fprintf(stderr, "%s: port2sock called with bogus argument %d", 
		progname, caller);
	term = 1;
	break;
  }; 
  
  if( i == 64 ) {
    fprintf(stderr, "%s: Problem in port2sock.\n", progname);
    term = 1;
  }
  
  return i;
}
*/


/*
 * seqmismatch prints warning messages and updates counters when sequence
 * numbers in packets are not as expected
 */
void
seqmismatch(unsigned expected, unsigned obtained)
{
  fprintf(stderr, "PARSE: WARNING: Sequence number mismatch. Expected"
	  " %u got %u.\n", expected, obtained);
#ifdef DEBUG
  fprintf(stdout, "PARSE: WARNING: Sequence number mismatch. Expected"
	  " %u got %u.\n", expected, obtained);
#endif
  if( expected < obtained ) {
    fprintf(stderr, "\tMissed %u bytes.\n", obtained - expected);
#ifdef DEBUG
    fprintf(stdout, "\tMissed %u bytes.\n", obtained - expected);
#endif
    lost_bytes += obtained - expected;
  }
  else {
    fprintf(stderr, "\tRepeated %u bytes.\n", expected - obtained);
#ifdef DEBUG
    fprintf(stdout, "\tRepeated %u bytes.\n", expected - obtained);
#endif
    repeated_bytes += expected - obtained;
  }
}


/* GWD
 * loopwrite stops the zr procees till it completes writing data.
 *
 */
int
loopwrite(int fd, const void *buf, size_t count, int pid)
{
  int ret, offset = 0, left = count;
  const int junksize = 65536;
  char junk[junksize];

  if( pid == 0 ) {
    fprintf(stderr, "LOOPWRITE: Writing too fast!\n");
    fprintf(stderr, "\tDon't have pid of zr process to stop it!\n");
    term = 1;
    return -1;
  }

  fprintf(stderr, "LOOPWRITE: Writing too fast! stopping zr pid %d\n", pid);
  kill(pid, SIGSTOP);

  fprintf(stderr, "LOOPWRITE: Sleeping for 1 second...\n");
  sleep(1);

  read(fd, junk, junksize);
  fprintf(stderr, "LOOPWRITE: Reading done.\n");

  while(left) {
    ret = write(fd, buf+offset, left);
    if(ret < 0) {
      if(errno != EAGAIN) {
	offset = -1;
	break;
      }
    } 
    else {
      fprintf(stderr, "LOOPWRITE: wrote %d bytes\n", ret);
      offset += ret;
      left -= ret;
    }
  }

  kill(pid, SIGCONT);

  return offset;
}

/*
 * mywrite calls write.
 * It calls loopwrite if write returns EAGAIN or if all the bytes are not
 * written. Any other errors are forwarded to caller.
 *
 */
int
mywrite(int fd, const void *buf, size_t count, int pid)
{
  int ret;

  ret = write(fd, buf, count); 

  if(ret < 0){
    if(errno == EAGAIN) {
      ret = loopwrite(fd, buf, count, pid);
    }
  }
  else if(ret != count) {
    fprintf(stderr, "MYWRITE: Tried to write %d bytes, wrote %d!\n", count,
	    ret);
    ret += loopwrite(fd, buf+ret, count-ret, pid);
  }

  return ret;
}


/*
 * parserecord understand the contents of the packet and does the needfull.
 * 
 * The frame_type and destination_port should always be as expected as the 
 * recording function takes care of this.
 * Assumptions are made about ihl, version, fragmentation and protocol fields
 * and a debugging message will be sent to stderr if they are broken.
 *
 * Optional TCP headers are not explicitely taken care of, but the correct TCP
 * data is found and sent using write.
 * Socket management is done when a syn, fin or res packet is encountered.
 * Otherwise if the packet contains TCP data, it is written to the appropriate
 * connection (if it exists). The pipe is always emptied by reading data from 
 * it before writing anything to it.
 */
void
parserecord(unsigned char *buf, int recordsize, int pid)
{
  unsigned char control;
  int urg, ack, pus, res, syn, fin;
  unsigned char tcp_header_len;
  int tcp_header_length;
  int sockno;
  const int opt = 1;
  struct sockaddr_in server;
#ifdef DEBUG
  static int i = 0;
#endif
  const int junksize = 65536;
  unsigned char junk[junksize];
  int readret;
  int close_connection = 0;
  int data_size;
  unsigned seq_no;

  // Skip MAC address of Destination[6].
  // Skip MAC address of Source[6].
  // Frame type[2] for IP frames.
  static unsigned char frame_type_ip[2] = {
    0x08, 0
  };
  // Version + Header Length[1], Type of Service[1].
  static unsigned char misc1[2] = {
    0x45, 0
  };
  // Skip Length[2], Id[2].
  // Fragmentation offset[2] for no fragmentation.
  static unsigned char frag[2] = {
    0x40, 0
  };
  // Skip Time to live[1].
  // Protocol[1] for TCP.
  static unsigned char proto[1] = {
    0x06
  };
  // Skip Checksum[2].
  // Skip IP address of Source[4].
  // Skip IP address of destination[4].
  // Skip Source port[2].
  // Destination port[2]. (6000)
  static unsigned char dst_port[2] = {
    0x17, 0x70
  };
  // Skip Sequence no[4], Acknowledgement no[4].

  if ( memcmp(buf + 12, frame_type_ip, 2) != 0
       || memcmp(buf + 36, dst_port, 2) != 0
       ) {
    // Panic.
    fprintf(stderr, "PARSE: Error in parsing headers\n");
    dump("error", buf, recordsize);
    term = 1;
    return;
  }

  if ( memcmp(buf + 14, misc1, 2) != 0
       || memcmp(buf + 20, frag, 2) != 0
       || memcmp(buf + 23, proto, 1) != 0
       ) {
    // Concern.
    fprintf(stderr, "PARSE: WARNING: Trouble in parsing headers\n");
    dump("warning", buf, recordsize);
    return;
  }

  // Extract Offset[4/8]
  tcp_header_len = (buf[46] & 0xf0) >> 4;
  tcp_header_length = tcp_header_len*4;
  data_size = recordsize-34-tcp_header_length;

  // Skip Reserved[6/8]
  // Extract Control Flags[6/8]
  control = buf[47];
  urg = (control & 0x20) >> 5;
  ack = (control & 0x10) >> 4;
  pus = (control & 0x08) >> 3;
  res = (control & 0x04) >> 2;
  syn = (control & 0x02) >> 1;
  fin = control & 0x01;

  seq_no = buf[38+0]*256u*256u*256u + buf[38+1]*256u*256u + buf[38+2]*256u 
           + buf[38+3];


  if(syn) {

#ifdef DEBUG
    fprintf(stdout, "PARSE: Starting syn\n");
    dump("syn", buf, recordsize);
#endif

    // Close socket if connection exists.
    sockno = port2sock(buf+34, 0);
    if ( sockno != -1 ) {
      close(connection[sockno].sock);
      port2sock(buf+34, 2);
    }

    // Prepare for new connection.
    sockno = port2sock(buf+34, 1);
    if ( sockno == -1 ) {
      fprintf(stderr, "PARSE: Exceeded max sockets\n");
      term = 1;
      return;
    }

    // Set sequence number to that of next packet to be sent on this
    // connection.
    connection[sockno].seq = seq_no + 1;

    // Create socket.
    connection[sockno].sock = socket( AF_INET, SOCK_STREAM, 0 );
    if (connection[sockno].sock == -1) {
      fprintf(stderr, "PARSE: Error in opening socket\n");
      term = 1;
    }
    
    // Configure the socket to send data immediately.
    if( setsockopt(connection[sockno].sock, IPPROTO_TCP, TCP_NODELAY,
		   (void *)&opt, sizeof(int)) == -1) {
      fprintf(stderr, "PARSE: Error in setting socket options\n");
      term = 1;
    }

    // Connect socket.
    server.sin_family = AF_INET;

    memcpy(&server.sin_addr.s_addr, buf+30, 4);
    memcpy(&server.sin_port, buf+36, 2);

#ifdef DEBUG
    //fprintf(stdout, "PARSE: Before connect\n");
#endif
    if (connect(connection[sockno].sock, (struct sockaddr *) &server,
		sizeof server) == -1) {
      fprintf(stderr, "PARSE: Error in connecting socket\n");
      fprintf(stderr, "\t Errno: %d\n", errno);
      term = 1;
    }

    fflush(NULL);

    // Make the socket non-blocking for recv calls.
    if( fcntl(connection[sockno].sock, F_SETFL, O_NONBLOCK) == -1 ) {
      fprintf(stderr, "PARSE: Error in making socket non-blocking\n");
      term = 1;
    }

#ifdef DEBUG
    //fprintf(stdout, "PARSE: Ending syn\n");
#endif
  }
  else if(fin) {

#ifdef DEBUG
    fprintf(stdout, "PARSE: Starting fin\n");
    dump("fin", buf, recordsize);
#endif

    // Close socket if connection exists.
    sockno = port2sock(buf+34, 0);
    if ( sockno != -1 ) {
      // Check if sequence number is as expected.
      if( connection[sockno].seq != seq_no ) {
	seqmismatch(connection[sockno].seq, seq_no );
      }
      close(connection[sockno].sock);
      port2sock(buf+34, 2);
    }
    else {
#ifdef DEBUG
      fprintf(stdout, "PARSE: No connection to close.\n");
#endif
    }

  }
  else if(res) {

#ifdef DEBUG
    fprintf(stdout, "PARSE: Connection reset by Peer\n");
    dump("reset", buf, recordsize);
#endif

    // Close socket if connection exists.
    sockno = port2sock(buf+34, 0);
    if ( sockno != -1 ) {
      // Check if sequence number is as expected.
      if( connection[sockno].seq != seq_no ) {
	seqmismatch(connection[sockno].seq, seq_no );
      }
      close(connection[sockno].sock);
      port2sock(buf+34, 2);
    }
    else {
#ifdef DEBUG
      fprintf(stdout, "PARSE: No connection to reset.\n");
#endif
    }

  }
  else {
	  
    if( data_size > 0 ) {

      sockno = port2sock(buf+34, 0);
      if ( sockno != -1 ) {
	do {
	  readret = read(connection[sockno].sock, junk, junksize);
#ifdef DEBUG
	  fprintf(stdout, "PARSE: Read returned  %d.\n", readret);
#endif
	  if( readret == 0 ) {
	    fprintf(stderr,"PARSE: Sender closed connection!\n");
#ifdef DEBUG
	    fprintf(stdout,"PARSE: Sender closed connection!\n");
#endif
	    close(connection[sockno].sock);
	    port2sock(buf+34, 2);
	    close_connection = 1;
	  }
	} while( readret == junksize );
#ifdef DEBUG
	fprintf(stdout, "PARSE: Before writing packet %d.\n", i);
	dump("write", buf, recordsize);
#endif
	if( close_connection == 0 ) {

	  // Check if sequence number is as expected.
	  if( connection[sockno].seq != seq_no ) {
	    seqmismatch(connection[sockno].seq, seq_no );
	  }

	  if(mywrite(connection[sockno].sock, buf+34+tcp_header_length,
		   data_size, pid) == -1) {
	    if( errno == EPIPE ) {
	      fprintf(stderr,
		      "PARSE: Got EPIPE in write, closing connection\n");
#ifdef DEBUG
	      fprintf(stdout,
		      "PARSE: Got EPIPE in write, closing connection\n");
#endif
	      close(connection[sockno].sock);
	      port2sock(buf+34, 2);
	    }
	    else {
	      fprintf(stderr, "PARSE: Error in write %s\n", strerror(errno));
	      term = 1;
	    }
	  }
	  fflush(NULL);

	  // Set sequence number to that of next packet to be sent on this
	  // connection.
	  connection[sockno].seq = seq_no + data_size; 

	}
	else {
	  fprintf(stdout, "PARSE: Connection closed. Not writing packet.\n");
	  dump("write", buf, recordsize);
	}
      }
      else {
	fprintf(stdout, "PARSE: No connection. Not writing packet.\n");
	dump("write", buf, recordsize);
      }
    }
    else {
#ifdef DEBUG
      fprintf(stdout, "PARSE: Packet without data portion.\n");
      dump("ignore", buf, recordsize);
#endif
    }
    
  }
#ifdef DEBUG
  fprintf(stdout, "PARSE: Record size: %4d   Data size: %4d\n", recordsize,
	  data_size);
  i++;
#endif
  fflush(NULL);
}

/*
void
parse_record(unsigned char *buf, int recordsize, int pid)
{
  unsigned char control;
  int urg, ack, pus, res, syn, fin;
  unsigned char tcp_header_len;
  int tcp_header_length;
  static int sock[64];
  int sockno;
  int close_socket = 0;
  const int opt = 1;
  struct sockaddr_in server;
#ifdef DEBUG
  static int i = 0;
#endif
  const int junksize = 4096;
  unsigned char junk[junksize];
  int readret;

  // Skip Destination MAC[6], Source MAC[6].
  // Frame type[2] for IP frames, Version + Header Length[1],
  // Type of Service[1].
  static unsigned char misc1[4] = {
    0x08, 0, 0x45, 0
  };
  // Skip Length[2], Id[2].
  // Fragmentation offset[2] for no fragmentation.
  static unsigned char frag[2] = {
    0x40, 0
  };
  // Skip Time to live[1].
  // Protocol[1] for TCP.
  static unsigned char proto[1] = {
    0x06
  };
  // Skip Checksum[2].
  // Skip Source IP[4], Destination IP[4]
  // Skip Source port[2].
  // Destination port[2]. (6000)
  static unsigned char dst_port[2] = {
    0x17, 0x70
  };
  // Skip Sequence no[4], Acknowledgement no[4].


  // Extract Offset[4/8]
  tcp_header_len = buf[46] & 0xf0;
  tcp_header_len /= 16;
  tcp_header_length = tcp_header_len*4; 
  //fprintf(stderr, "%s: TCP Header Length %d\n", progname, tcp_header_length);

  // Skip Reserved[6/8]
  // Extract Control Flags[6/8]
  control = buf[47];
  urg = control & 0x20;
  ack = control & 0x10;
  pus = control & 0x08;
  res = control & 0x04;
  syn = control & 0x02;
  fin = control & 0x01;

  // We tolerate fragemntation if we don't expect any data --
  // and we don't expecet any for res,syn, and fin packets

  if ( memcmp(buf + 12, misc1, 4) != 0 
	|| ( (memcmp(buf + 20, frag, 2) != 0) 
	      && !(res | syn | fin)  )
	|| memcmp(buf + 23, proto, 1) != 0 
	|| memcmp(buf + 36, dst_port, 2) != 0 ) {
    // Panic.
    fprintf(stderr, "%s: Error in parsing headers\n", progname);
    dump("parserecord", buf, recordsize);
    term = 1;
    return;
  }


  if(syn) {
#ifdef DEBUG
    fprintf(stderr, "%s: Starting syn\n", progname);
#endif

    // Get the mapping from source port to socket.
    sockno = port2sock(buf+34, 0, &close_socket);
    
    if( close_socket == 1 ) {
      close(sock[sockno]);
      close_socket = 0;
    }
    // Create socket.
    sock[sockno] = socket( AF_INET, SOCK_STREAM, 0 );
    if (sock[sockno] == -1) {
      fprintf(stderr, "%s: Error in opening socket: %s\n", progname,
	      strerror(errno));
      term = 1;
    }
    
    // Configure the socket to send data immediately.
    if( setsockopt(sock[sockno], IPPROTO_TCP, TCP_NODELAY, (void *)&opt,
		   sizeof(int)) == -1) {
      fprintf(stderr, "%s: Error in setting socket options: %s\n", progname,
	      strerror(errno));
      term = 1;
    }

    // Connect socket.
    server.sin_family = AF_INET;
    memcpy(&server.sin_addr.s_addr, buf+30, 4);
    memcpy(&server.sin_port, buf+36, 2);

#ifdef DEBUG
    //fprintf(stderr, "%s: Before connect\n", progname);
#endif
    if (connect(sock[sockno], (struct sockaddr *) &server, sizeof server)
	== -1) {
      fprintf(stderr, "%s: Error in connecting socket: %s\n", progname,
	      strerror(errno));
      term = 1;
    }

    fflush(NULL);

    // Make the socket non-blocking for recv calls.
    if( fcntl(sock[sockno], F_SETFL, O_NONBLOCK) == -1 ) {
      fprintf(stderr, "%s: Error in making socket non-blocking: %s\n",
	      progname, strerror(errno));
      term = 1;
    }
  }
  else if(fin) {
#ifdef DEBUG
    fprintf(stderr, "%s: Starting fin\n", progname);
#endif

    // Get the mapping from source port to socket.
    sockno = port2sock(buf+34, 1, &close_socket);

    close(sock[sockno]);
    //term = 1;
  }
  else if(res) {
#ifdef DEBUG
    fprintf(stderr, "%s: Connection reset by Peer\n", progname);
    dump("reset", buf, recordsize);
#endif

    // Get the mapping from source port to socket.
    sockno = port2sock(buf+34, 2, &close_socket);

    close(sock[sockno]);
    //term = 1;
  }
  else {
    if(recordsize-34-tcp_header_length > 0 ) {
#ifdef DEBUG
      i++;
      fprintf(stderr, "%s: Before writing packet %d.\n", progname, i);
#endif

      // Get the mapping from source port to socket.
      sockno = port2sock(buf+34, 3, &close_socket);

      if(mywrite(sock[sockno], buf+34+tcp_header_length , 
	       recordsize-34-tcp_header_length, pid) == -1) {
	switch(errno) {

	  //	  case EAGAIN:
	  //fprintf(stderr, "Writing too fast! Stopping zr process, pid %d\n",
	  //		  pid);
	  //	   kill(pid, SIGSTOP);
	  //	   while(1) {
	  //	    if(write(sock[sockno], buf+34+tcp_header_length , 
	  //		     recordsize-34-tcp_header_length) >=0) {
	  //fprintf(stderr, "Write succeeded! Continuing zr process\n");
	  //	      kill(pid, SIGCONT);
	  //	      break;
	  //	    } else if(errno != EAGAIN) {
	  //	      fprintf(stderr, "Error in write: %s\n", strerror(errno));
	  //	      term=1;
	  //	      break;
	  //	    }
	  //	   }
	  //	   break;

	 case EPIPE:
		fprintf(stderr, "Got EPIPE.  Closing connection\n");
		close(sock[sockno]);
		port2sock(buf+34, 4, &close_socket);
	   	break;
	 default: 
		fprintf(stderr, "%s: Error in write: %s\n", progname, 
		        strerror(errno));
		term = 1;
	}
      }
      fflush(NULL);
      readret = 0;
      do {
	readret = read(sock[sockno], junk, junksize);
#ifdef DEBUG
	fprintf(stderr, "%s: Read returned  %d.\n", progname, readret);
#endif
      } while( readret == 4096 );
    }
  }
#ifdef DEBUG
  fprintf(stderr, "%s: Parse record size: %d\n", progname, recordsize);
  dump("parserecord", buf, recordsize);
#endif
  fflush(NULL);
}
*/

int
main(int argc, char *argv[])
{
  struct sigaction sa;
  int phys_fd;
  int virt_fd;
  char buf[1024*64];
  int len;
  int ret;
  int log_rep; // 0 for log, 1 for rep.
  unsigned char ip[4] = { 0, 0, 0, 0 };
  int filter_pid = 0;
  
  progname = *argv;
  
  if (argc < 3) {
    fprintf(stderr, "Usage: %s -l <GuestIP> [filter]\n", argv[0]);
    fprintf(stderr, " OR  : %s -r <XServerIP> <pid>\n", argv[0]);
    exit(1);
  }

  if ( strcmp(argv[1], "-l") == 0 )
    log_rep = 0;
  else if ( strcmp(argv[1], "-r") == 0 )
    log_rep = 1;
  else {
    fprintf(stderr, "Usage: %s -l <GuestIP> [filter]\n", argv[0]);
    fprintf(stderr, " OR  : %s -r <XServerIP> <pid>\n", argv[0]);
    exit(1);
  }

  ret = inet_aton(argv[2], (struct in_addr *)ip );
  assert(0 <= ret);

  if(argc > 3)
    filter_pid = atoi(argv[3]);

#ifdef DEBUG
  fprintf(stdout, "IP Address: ");
  fprintf(stdout, "%02x %02x %02x %02x\n", ip[0], ip[1], ip[2], ip[3]);
#endif

  config_read();
  
  sa.sa_handler = int_handler;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  ret = sigaction(SIGINT, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGQUIT, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGHUP, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGPIPE, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  
  virt_fd = virt_open();
  phys_fd = phys_open();
  
  while (! term) {
    fd_set rfds;
    fd_set wfds;
    fd_set efds;
    int maxfd;
    
    /*
     * wait for message
     */
    maxfd = -1;
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_ZERO(&efds);
    if (0 <= phys_fd) {
      FD_SET(phys_fd, &rfds);
      if (maxfd < phys_fd) maxfd = phys_fd;
    }
    if (0 <= virt_fd) {
      FD_SET(virt_fd, &rfds);
      if (maxfd < virt_fd) maxfd = virt_fd;
    }
    ret = select(maxfd + 1, &rfds, &wfds, &efds, NULL);
    if (ret < 0 && errno != EINTR) {
      perror("select");
      exit(1);
    }
    
    if (0 <= phys_fd && 0 < ret && FD_ISSET(phys_fd, &rfds)) {
      /* message from physical net received */
      len = phys_recv(phys_fd, buf, sizeof(buf));
      if (len < 0) {
	fprintf(stderr, "%s: phys_recv(%d, %p %d): %s.\n",
		progname, phys_fd, buf, sizeof(buf),
		strerror(errno));
      } else {
#ifdef DEBUG
	dump("p->v", buf, len);
#endif
	if ( (log_rep == 0) && (!to_guest_filter(buf, len, ip, filter_pid)) ) {
	  ret = virt_send(virt_fd, buf, len);
	  if (ret < 0) {
	    fprintf(stderr, "%s: virt_send: %s.\n",
		    progname,
		    strerror(errno));
	  }
	}
      }
    }
    if (0 <= virt_fd && 0 < ret && FD_ISSET(virt_fd, &rfds)) {
      /* message from virtual net received */
      len = virt_recv(virt_fd, buf, sizeof(buf));
      if (len < 0) {
	fprintf(stderr, "%s: virt_recv: %s.\n",
		progname,
		strerror(errno));
      } else {
#ifdef DEBUG
	dump("v->p", buf, len);
#endif
	if ( log_rep == 0) {
	  ret = phys_send(phys_fd, buf, len);
	  if (ret < 0) {
	    fprintf(stderr, "%s: phys_send: %s.\n",
		    progname,
		    strerror(errno));
	  }
	}
	else if (!to_xserver_filter(buf, len, ip)) {
	  parserecord(buf, len, filter_pid);
	}
      }
    }
  }
  
  phys_close(phys_fd);
  virt_close(virt_fd);

  if( lost_bytes > 0 ) {
    fprintf(stderr, "Lost %u bytes totally.\n", lost_bytes);
#ifdef DEBUG
    fprintf(stdout, "Lost %u bytes totally.\n", lost_bytes);
#endif
  }
  if( repeated_bytes > 0 ) {
    fprintf(stderr, "Repeated %u bytes totally.\n", repeated_bytes);
#ifdef DEBUG
    fprintf(stdout, "Repeated %u bytes totally.\n", repeated_bytes);
#endif
  }
    
  return 0;
}
