/*
 * record.c
 *
 * Records all packets going to X Server (port 6000) on some IP address.
 * Uses raw socket connection to get the packets.
 * Stores packet sizes in recordsize and packet data in recorddata in current
 * directory.
 *
 * PROBLEM: Will miss packets if they come too fast.
 *
 */

/* Hack to get ntohl etc working */
#define _LINUX_BYTEORDER_GENERIC_H	1

#define _GNU_SOURCE

#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <net/ethernet.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netpacket/packet.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

//#define DEBUG
//#define DEBUG1 // To really debug. WARNING: May cause packets to be dropped.


char *progname;

volatile int term = 0;
struct ifreq saved_ifr;

FILE *fp1, *fp2;

int eth;


void
int_handler(int sig)
{
  fprintf(stderr, "SIGNAL: going down on signal %d : %s\n", sig,
	  strsignal(sig));
  term = 1;
}


void
openfile(void)
{
  if( (fp1 = fopen("./recordsize", "w")) == NULL ) {
    fprintf(stderr, "FILEIO: Unable to open file ./recordsize\n");
    term = 1;
  }
  if( (fp2 = fopen("./recorddata", "w")) == NULL ) {
    fprintf(stderr, "FILEIO: Unable to open file ./recorddata\n");
    term = 1;
  }
}


void
closefile(void)
{
  fclose(fp1);
  fclose(fp2);
}


void
record(unsigned char*buf, int bufsize)
{
  char buffersize[16];
  sprintf(buffersize, "%d ", bufsize);
  if( (fwrite((void *)buffersize, 1, strlen(buffersize), fp1)) 
      != strlen(buffersize) ){
    fprintf(stderr, "FILEIO: Error in writing to file ./recordsize\n");
    fclose(fp1);
    term = 1;
  }

  if( (fwrite((void *)buf, 1, bufsize, fp2)) 
      != bufsize ){
    fprintf(stderr, "FILEIO: Error in writing to file ./recorddata\n");
    fclose(fp2);
    term = 1;
  }
}


void
dump(char *string, unsigned char *buf, int bufsize)
{
  int i;

  fprintf(stdout, "%s: \n", string);
  for (i = 0; i < bufsize && i < 100; i++) {
    fprintf(stdout, "%02x ", buf[i]);
  }
  fprintf(stdout, "\n");
}


int
phys_open(void)
{
  int fd;
  struct ifreq ifr;
  struct sockaddr_ll addr;
  int ifindex;
  int ret;
  
  
  /* ================= */
  /* create raw socket */
  /* ================= */
  
  fd = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
  if (fd < 0) {
    fprintf(stderr, "%s: socket(PF_PACKET, SOCK_RAW, ETH_P_ALL): %s.\n",
	    progname, strerror(errno));
    return -1;
  }
  
  /* =================== */
  /* get interface index */
  /* =================== */
  
  memset(&ifr, 0, sizeof(ifr));
  sprintf(ifr.ifr_name, "eth%d", eth);
  ret = ioctl(fd, SIOCGIFINDEX, &ifr);
  if (ret < 0) {
    fprintf(stderr, "%s: ioctl(..., SIOCGIFINDEX, ...): %s.\n",
	    progname, strerror(errno));
    (void) close(fd);
    return -1;
  }
  
  ifindex = ifr.ifr_ifindex;

  /* ================== */
  /* set promiscue mode */
  /* ================== */
  
  memset(&ifr, 0, sizeof(ifr));
  sprintf(ifr.ifr_name, "eth%d", eth);
  ret = ioctl(fd, SIOCGIFFLAGS, &ifr);
  if (ret < 0) {
    fprintf(stderr, "%s: ioctl(..., SIOCGIFFLAGS, ...): %s.\n",
	    progname, strerror(errno));
    (void) close(fd);
    return -1;
  }
  saved_ifr = ifr;
  ifr.ifr_flags |= IFF_PROMISC;
  ret = ioctl(fd, SIOCSIFFLAGS, &ifr);
  if (ret < 0) {
    fprintf(stderr, "%s: ioctl(..., SIOCSIFFLAGS, ...): %s.\n",
	    progname, strerror(errno));
    (void) close(fd);
    return -1;
  }
  
  /* =========== */
  /* bind socket */
  /* =========== */
  
  memset(&addr, 0, sizeof(addr));
  addr.sll_family = PF_PACKET;
  addr.sll_protocol = htons(ETH_P_ALL);
  addr.sll_ifindex = ifindex;
  ret = bind(fd, (struct sockaddr *) &addr, sizeof(addr));
  if (ret < 0) {
    fprintf(stderr, "%s: bind: %s.\n",
	    progname, strerror(errno));
    (void) close(fd);
    return -1;
  }

#ifdef DEBUG  
  fprintf(stdout, "ifindex: %d\n", ifindex);
  fprintf(stdout, "phys_open: fd = %d\n", fd);
#endif

  return fd;
}


int
phys_recv(int fd, unsigned char *buf, int bufsize)
{
  struct sockaddr_ll addr;
  int addrlen;
  int len;

  addrlen = sizeof(addr);  
  len = recvfrom(fd, buf, bufsize, 0, (struct sockaddr *) &addr, &addrlen);

  return len;
}


int
phys_close(int fd)
{
  int ret;
  
  if (fd < 0) {
    return 0;
  }
  
  ret = ioctl(fd, SIOCSIFFLAGS, &saved_ifr);
  if (ret < 0) {
    fprintf(stderr, "%s: ioctl(..., SIOCSIFFLAGS, ...): %s.\n",
	    progname, strerror(errno));
  }
  
  ret = close(fd);
  
#ifdef DEBUG
  fprintf(stdout, "phys_close: ret = %d\n", ret);
#endif

  return ret;
}


/* 
 * Defining our own filter function.
 * Only packets destined for the guest X Server will be recorded.
 * Return 0 if packet is to be forwarded to guest,
 * 1 if it needs to be dropped.
 */
int
filter(unsigned char *buf, int bufsize, unsigned char *ip)
{

  // Frame type for IP[2].
  static unsigned char frame_type_ip[2] = {
    0x08, 0
  };

  // Destination port for X Server[2]. (6000)
  static unsigned char dst_port_x[2] = {
    0x17, 0x70
  };


  if( (memcmp(buf + 12, frame_type_ip, 2) == 0) && 
      (memcmp(buf + 30, ip, 4) == 0) &&
      (memcmp(buf + 36, dst_port_x, 2) == 0) ) {
    return 0;
  }

  return 1;
}


int
main(int argc, char **argv)
{
  struct sigaction sa;
  int phys_fd;
  char buf[1024*64];
  int len;
  int ret;
  unsigned char ip[4] = { 0, 0, 0, 0 };
  
  progname = *argv;
  
  sa.sa_handler = int_handler;
  sa.sa_flags = 0;
  sigemptyset(&sa.sa_mask);
  ret = sigaction(SIGINT, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGQUIT, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGHUP, &sa, (struct sigaction *) 0);
  assert(0 <= ret);
  ret = sigaction(SIGPIPE, &sa, (struct sigaction *) 0);
  assert(0 <= ret);

  if (argc < 2) {
    fprintf(stderr, "Usage: %s <IP>\n", argv[0]);
    exit(1);
  }

  ret = inet_aton(argv[1], (struct in_addr *)ip );
  assert(0 <= ret);

#ifdef DEBUG
  fprintf(stdout, "Recording packets going to XServer (6000) on IP: ");
  fprintf(stdout, "%02x %02x %02x %02x\n", ip[0], ip[1], ip[2], ip[3]);
#endif
  
  phys_fd = phys_open();
  openfile();
  
  while (! term) {
    fd_set rfds;
    fd_set wfds;
    fd_set efds;
    int maxfd;
    
    /*
     * wait for message
     */
    maxfd = -1;
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_ZERO(&efds);
    if (0 <= phys_fd) {
      FD_SET(phys_fd, &rfds);
      if (maxfd < phys_fd) maxfd = phys_fd;
    }
    ret = select(maxfd + 1, &rfds, &wfds, &efds, NULL);
    if (ret < 0 && errno != EINTR) {
      perror("select");
      exit(1);
    }
    
    if (0 <= phys_fd && 0 < ret && FD_ISSET(phys_fd, &rfds)) {
      /* message from physical net received */
      len = phys_recv(phys_fd, buf, sizeof(buf));
      if (len < 0) {
	fprintf(stderr, "%s: phys_recv: %s.\n",
		progname,
		strerror(errno));
      } else {
	if (! filter(buf, len, ip)) {
#ifdef DEBUG1
	  dump("Recieved", buf, len);
#endif
	  record(buf, len);
	}
      }
    }
  }
  
  closefile();
  phys_close(phys_fd);
  
  return 0;
}
