-- $Id: single-node.vhdl,v 1.1.1.1 2003/05/27 21:45:34 root Exp $
-- vim:tabstop=8:shiftwidth=8:textwidth=72

-- VHDL comment character is --, comments are until end of line
--
-- VHDL-description of a single virtual UMLinux node
--

-- node-internal things are prefixed with int_ in this example.
-- I have tried to avoid using names which are identical to keywords or
-- predefined components, so it is clear, which names can be freely
-- chosen and which cannot.
-- node-external things are prefixed with ext_. node-external things in
-- this case include interfaces, which other nodes can access.

-- ####################################################################
-- define an entity named 'node'
-- ####################################################################
-- define interface of entity 'node'
entity node is
	-- variable definitions, used later on.
	-- the 'generic' statement includes (in parentheses) a list of
	-- variable definitions separated by semicolons. name of
	-- variable, type and value are given.
	generic(
		memsize : integer := 128;
		disksize : integer := 1
	);
	-- the interface (port) of this entity consists of its network
	-- interface "ext_eth0" which is connected to the ethernet 'ethbus'
	-- you could use this to connect several nodes to an ethernet
	-- cable.
	-- the 'port' statement includes (in parentheses) a list of
	-- port-names (e.g. "ext_eth0") and their types (e.g.
	-- "inout ethbus"). There is a colon between port-name and type.
	-- the port-name/type pairs are separated by semicolons.
	port(
		ext_eth0 : inout ethbus
		-- add more ports here (e.g. if you have more than one
		-- ethernet-card you should probably have a separate
		-- port here for each card. you'll also need a semicolon
		-- after the first card (and don't put a semicolon after
		-- the last definition before the closing parenthesis!)
		-- ext_eth1 : inout ethbus;
		-- ext_eth2 : inout ethbus
	);
end node;

-- define structural internals of entity 'node'
-- use whatever you want instead of "structural", it is ignored anyway
architecture structural of node is
	-- === internal busses ========================================
	-- there shouldn't be any need to change this definition of the
	-- busses

	-- for UMLinux we use signals to model busses. the elements in
	-- the list are separated by semicolons. for each signal the
	-- name (e.g. int_mem) and the type (e.g. membus) must be given.

	-- define frontside bus. type membus is predefined.
	-- cpu, memory and mem2isa-bridge will be connected to this
	signal	int_mem : membus;

	-- define isa bus. type isabus is predefined.
	-- mem2isa-bridge, idectrl and ethcard will be connected to this
	signal	int_isa : isabus;

	-- define ide bus. type idebus is predefined.
	-- idectrl and idedisk will be connected to this
	signal	int_ide : idebus;
	-- === end of internal busses definition ======================
begin
	-- === internal components ====================================
	-- port map tells you which port of the (predefined) component's
	-- interface is connected to which bus. port- and bus-type must
	-- be the same (i.e. you cannot connect a port of type idebus to
	-- a bus of type ethbus).
	-- cpus are predefined components with a single interface/port
	-- named 'mem'. this interface can be connected to a bus of type
	-- 'membus' (in our case we connect it to the frontside-bus
	-- (int_mem)).
	cpu0 : cpu port map( mem => int_mem );
	-- add more cpu's here:
	-- cpu1 : cpu port map( mem => int_mem );
	-- cpu2 : cpu port map( mem => int_mem );
	-- ...

	-- there is only one memory.
	-- the predefined memory-component is configurable via the
	-- attribute 'size', which is, in this case, set to whatever
	-- value the variable memsize (defined above) has.
	mem0 : memory generic map( size => memsize )
	-- memory is connected to the frontside-bus (int_mem)
			port map( mem => int_mem );
	-- the ethernet card "eth0" is connected to internal isabus and
	-- external ethernet

	-- the north-/southbridge are collapsed into a single component
	-- which connects the membus and the isabus.
	-- no need to change this.
	northsouth : mem2isa port map( mem => int_mem, isa => int_isa );


	-- ethernet card is connected to internal isa bus and external
	-- ethernet bus.
	eth0 : ethcard port map( isa => int_isa, eth => ext_eth0 );
	-- add more network cards here:
	-- eth1 : ethcard port map( isa => int_isa, eth => ext_eth1 );
	-- eth2 : ethcard port map( isa => int_isa, eth => ext_eth2 );

	-- ide-controller is needed to access idedisks and is connected
	-- to  internal isa- and ide-busses.
	idectrl0 : idectrl port map( isa => int_isa, ide => int_ide );

	-- predefined idedisk-components have a size-attribute which is
	-- set to the value of the disksize variable defined above.
	-- harddisks and cdroms are connected to the internal ide-bus.
	-- names of disks given here are ignored! Names are generated
	-- automatically starting with hda, hdb, hdc ...
	hd0  : idedisk
		generic map ( size => disksize )
		port map( ide => int_ide );
	-- define more harddisks here:
	-- hd1  : idedisk
	--	generic map ( size => disksize )
	--	port map( ide => int_ide );

	-- no cdrom yet!
	-- cd0 : idecdrom
	-- 	port map( ide => int_ide );
	-- === end of internal components definition ==================
end structural;
-- ####################################################################
-- end of definition of entity 'node'
-- ####################################################################
